from typing import List
from sqlmodel import Session
from datetime import datetime, timezone
from obspy import UTCDateTime

from geomagio.api.db.models.metadata import (
    MetadataTable,
    Metadata,
    MetadataQuery,
    MetadataUpdate,
)
from geomagio.api.db.models.metadata_history import MetadataHistoryTable
from geomagio.api.db.factories import db_helpers


class MetadataDatabaseFactory:
    def __init__(self, engine):
        self.engine = engine

    async def create_metadata(self, new_metadata: Metadata) -> Metadata:
        with Session(self.engine) as db_session:
            # convert Metadata input into MetadataTable object
            new_metadata_values = new_metadata.model_dump(by_alias=True)
            new_metadata_table_model = MetadataTable(**new_metadata_values)

            new_metadata_table_model = MetadataTable.model_validate(
                new_metadata_table_model
            )
            new_metadata_table_model.id = None
            if new_metadata_table_model.created_time is None:
                new_metadata_table_model.created_time = UTCDateTime.now()

            # convert UTCDateTime to aware datetime for sqlalchemy
            new_metadata_table_model = db_helpers.convert_utcdatetime_to_datetime(
                metadata=new_metadata_table_model
            )

            db_session.add(new_metadata_table_model)
            db_session.commit()
            db_session.refresh(new_metadata_table_model)

            # convert newly created metadata row into Metadata object
            newly_created_metadata_values = new_metadata_table_model.model_dump(
                by_alias=True, exclude_none=True
            )
            returned_metadata = Metadata(**newly_created_metadata_values)

            return returned_metadata

    async def get_metadata(self, query: MetadataQuery) -> List[Metadata]:
        with Session(self.engine) as db_session:
            query = MetadataQuery.model_validate(query)
            statement = db_helpers.build_metadata_search_query(
                query=query, table=MetadataTable
            )

            results = db_session.exec(statement)

            returned_metadata = []

            for result in results:
                found_metadata_values = result.model_dump(
                    by_alias=True, exclude_none=True
                )
                found_metadata = Metadata(**found_metadata_values)
                returned_metadata.append(found_metadata)

            return returned_metadata

    async def get_metadata_by_id(self, id=int) -> Metadata:
        if type(id) != int:
            raise ValueError("expected an int")

        with Session(self.engine) as db_session:
            result = db_session.get(MetadataTable, id)
            if result != None:
                result_values = result.model_dump(by_alias=True, exclude_none=True)
                return Metadata(**result_values)
            else:
                return None

    async def update_metadata(
        self, updated_metadata: MetadataUpdate, updated_by: str
    ) -> Metadata:
        with Session(self.engine) as db_session:
            # get original metadata
            original_metadata = db_session.get(MetadataTable, updated_metadata.id)
            if original_metadata is None:
                raise ValueError("metadata not found")

            # save original metadata as a new metadata_history record
            new_metadata_history_values = original_metadata.model_dump()
            new_metadata_history = MetadataHistoryTable(**new_metadata_history_values)

            # remove the id and set metadata_id to the original_metadata id
            new_metadata_history.id = None
            new_metadata_history.metadata_id = original_metadata.id

            db_session.add(new_metadata_history)
            db_session.flush()

            # set updated values on original_metadata
            # do not model_validate the inputs otherwise the defaults will all be set to None.
            # sqlmodel_update removes any values set to None
            update_values = updated_metadata.model_dump(
                exclude_unset=True, by_alias=True
            )
            original_metadata.sqlmodel_update(update_values)

            original_metadata.updated_by = updated_by
            original_metadata.updated_time = datetime.now(timezone.utc)
            db_session.add(original_metadata)
            db_session.commit()
            db_session.refresh(original_metadata)

            # convert newly updated metadata row to Metadata type
            return_metadata_values = original_metadata.model_dump(
                by_alias=True, exclude_none=True
            )
            return Metadata(**return_metadata_values)

    async def batch_create_metadata(self, new_metadatas: List[Metadata]):
        with Session(self.engine) as db_session:
            for new_metadata in new_metadatas:
                new_metadata_values = new_metadata.model_dump(by_alias=True)
                new_metadata_table_model = MetadataTable(**new_metadata_values)

                new_metadata_table_model = MetadataTable.model_validate(
                    new_metadata_table_model
                )
                new_metadata_table_model.id = None
                if new_metadata_table_model.created_time is None:
                    new_metadata_table_model.created_time = UTCDateTime.now()

                # convert UTCDateTime to aware datetime for sqlalchemy
                new_metadata_table_model = db_helpers.convert_utcdatetime_to_datetime(
                    metadata=new_metadata_table_model
                )

                db_session.add(new_metadata_table_model)
            db_session.commit()
            db_session.refresh(new_metadata_table_model)
        return new_metadatas
