from typing import List
from sqlmodel import Session, select

from geomagio.api.db.factories import db_helpers
from geomagio.api.db.models.metadata import Metadata, MetadataTable, MetadataQuery
from geomagio.api.db.models.metadata_history import (
    MetadataHistoryTable,
    MetadataHistory,
)


class MetadataHistoryDatabaseFactory:
    def __init__(self, engine):
        self.engine = engine

    async def get_metadata_history_by_id(self, id=int) -> MetadataHistory:
        if type(id) != int:
            raise ValueError("expected an int")

        with Session(self.engine) as db_session:
            result = db_session.get(MetadataHistoryTable, id)
            if result != None:
                result_values = result.model_dump(by_alias=True)
                return MetadataHistory(**result_values)
            else:
                return None

    async def get_metadata_history(self, query: MetadataQuery) -> List[MetadataHistory]:
        with Session(self.engine) as db_session:
            query = MetadataQuery.model_validate(query)
            statement = db_helpers.build_metadata_search_query(
                query=query, table=MetadataHistoryTable
            )

            results = db_session.exec(statement)

            returned_results = []
            for result in results:
                result_values = result.model_dump(by_alias=True)
                returned_results.append(MetadataHistory(**result_values))

            return returned_results

    async def get_metadata_history_by_metadata_id(self, metadata_id: int):
        if type(metadata_id) != int:
            raise ValueError("expected an int")

        returned_list = []
        with Session(self.engine) as db_session:
            statement = (
                select(MetadataHistoryTable, MetadataTable)
                .join(MetadataTable)
                .where(MetadataTable.id == metadata_id)
            )
            results = db_session.exec(statement)
            history = results.all()

            # if there's no metadata_history associated with the metdata object, just return the metadata ojbect
            if len(history) == 0:
                metadata = db_session.get(MetadataTable, metadata_id)
                if metadata == None:
                    return None
                else:
                    result_values = metadata.model_dump(by_alias=True)
                    return [Metadata(**result_values)]

            metadata_list = []
            for metadata_history, metadata in history:
                metadata_history_values = metadata_history.model_dump(by_alias=True)
                formatted_metadata_history = MetadataHistory(**metadata_history_values)
                returned_list.insert(0, formatted_metadata_history)

                # make sure only 1 metadata is appended. if not, raise a big error
                if len(metadata_list) == 0:
                    metadata_list.append(metadata)
                else:
                    if metadata_list[0] != metadata:
                        raise SystemError(
                            "get_metadata_history_by_metadata_id returned more than one metadata for id: ",
                            metadata_id,
                        )

            # insert Metadata object to the front of the list
            formatted_metadata_values = metadata_list[0].model_dump(by_alias=True)
            formatted_metadata = Metadata(**formatted_metadata_values)
            returned_list.insert(0, formatted_metadata)

            return returned_list

    async def batch_create_metadata_history(self, new_metadatas: List[MetadataHistory]):
        with Session(self.engine) as db_session:
            for new_metadata in new_metadatas:
                new_metadata = MetadataHistoryTable.model_validate(new_metadata)
                new_metadata.id = None

                # convert UTCDateTime to aware datetime for sqlalchemy
                new_metadata = db_helpers.convert_utcdatetime_to_datetime(
                    metadata=new_metadata
                )

                db_session.add(new_metadata)
            db_session.commit()
