import os
import json
import uuid

from typing import Mapping
from datetime import datetime, timezone
from sqlmodel import Session, select

from geomagio.api.db.models.session import session
from geomagio.api.secure.encryption import get_fernet


class SessionDatabaseFactory:
    def __init__(self, engine):
        self.encryption = get_fernet()
        self.engine = engine

    async def get_session(self, session_id: str) -> session:
        with Session(self.engine) as db_session:
            statement = select(session).where(session.session_id == session_id)
            results = db_session.exec(statement)
            existing_session = results.one()

            # if existing_session is None:
            #     return None

            data = self.encryption.decrypt(existing_session.data.encode("utf8"))
            return json.loads(data)

    async def save_session(self, session_id: str, data: Mapping) -> None:
        updated = datetime.now(timezone.utc)

        data = json.dumps(data)
        data = self.encryption.encrypt(data.encode("utf8")).decode("utf8")

        with Session(self.engine) as db_session:
            # check if session exists before updating
            statement = select(session).where(session.session_id == session_id)
            results = db_session.exec(statement)
            existing_session = results.first()

            if not existing_session:
                new_session = session(session_id=session_id, data=data, updated=updated)
                db_session.add(new_session)
            else:
                existing_session.updated = updated
                existing_session.data = data
                db_session.add(existing_session)

            db_session.commit()

    async def delete_session(self, session_id: str) -> None:
        with Session(self.engine) as db_session:
            statement = select(session).where(session.session_id == session_id)
            results = db_session.exec(statement)
            existing_session = results.first()

            if existing_session is None:
                return None

            db_session.delete(existing_session)
            db_session.commit()
