from datetime import timezone
from typing import Any
from sqlmodel import select, or_

from geomagio.api.db.models.metadata import MetadataQuery
from geomagio.api.db.database import Tables


def convert_utcdatetime_to_datetime(metadata: Any):
    if metadata.created_time is not None:
        metadata.created_time = metadata.created_time.datetime.replace(
            tzinfo=timezone.utc
        )
    if metadata.updated_time is not None:
        metadata.updated_time = metadata.updated_time.datetime.replace(
            tzinfo=timezone.utc
        )
    if metadata.starttime is not None:
        metadata.starttime = metadata.starttime.datetime.replace(tzinfo=timezone.utc)
    if metadata.endtime is not None:
        metadata.endtime = metadata.endtime.datetime.replace(tzinfo=timezone.utc)
    return metadata


def build_metadata_search_query(query: MetadataQuery, table: Tables):
    # convert UTCDateTime to aware datetime for sqlalchemy
    if query.created_before is not None:
        query.created_before = query.created_before.datetime.replace(
            tzinfo=timezone.utc
        )
    if query.created_after is not None:
        query.created_after = query.created_after.datetime.replace(tzinfo=timezone.utc)
    if query.starttime is not None:
        query.starttime = query.starttime.datetime.replace(tzinfo=timezone.utc)
    if query.endtime is not None:
        query.endtime = query.endtime.datetime.replace(tzinfo=timezone.utc)

    # build select statement
    statement = select(table)
    if query.id:
        statement = statement.where(table.id == query.id)
    if query.category:
        statement = statement.where(table.category == query.category)
    if query.network:
        statement = statement.where(table.network == query.network)
    if query.station:
        statement = statement.where(table.station == query.station)
    if query.created_by:
        statement = statement.where(table.created_by == query.created_by)

    # using "like" to search for channels and locations to account for potential inconsistencies over time
    if query.channel:
        statement = statement.where(table.channel.like(query.channel))
    if query.location:
        statement = statement.where(table.location.like(query.location))

    if query.starttime:
        statement = statement.where(
            or_(
                table.endtime == None,
                table.endtime > query.starttime,
            )
        )
    if query.endtime:
        statement = statement.where(
            or_(
                table.starttime == None,
                table.starttime < query.endtime,
            )
        )
    if query.created_after:
        statement = statement.where(table.created_time > query.created_after)
    if query.created_before:
        statement = statement.where(table.created_time < query.created_before)
    if query.data_valid is not None:
        statement = statement.where(table.data_valid == query.data_valid)

    # users can search for multiple statuses
    if query.status is not None:
        statement = statement.where(table.status.in_(query.status))

    return statement
