import sqlalchemy_utc

from enum import Enum
from obspy import UTCDateTime
from datetime import datetime, timezone
from typing import Dict, List, Any
from pydantic import BaseModel, field_serializer
from sqlmodel import SQLModel, Field, Column, Text, JSON, Index

from geomagio.pydantic_utcdatetime import CustomUTCDateTimeType


class MetadataCategory(str, Enum):
    ADJUSTED_MATRIX = "adjusted-matrix"
    FLAG = "flag"
    INSTRUMENT = "instrument"
    OBSERVATORY = "observatory"
    READING = "reading"


class MetadataTable(SQLModel, table=True):
    __tablename__ = "metadata"

    id: int | None = Field(primary_key=True)
    created_by: str | None = Field(index=True)
    created_time: CustomUTCDateTimeType | None = Field(
        default=UTCDateTime.now(),
        sa_column=Column(sqlalchemy_utc.UtcDateTime, index=True),
    )
    updated_by: str | None = Field(index=True)
    updated_time: CustomUTCDateTimeType | None = Field(
        sa_column=Column(sqlalchemy_utc.UtcDateTime, index=True)
    )
    starttime: CustomUTCDateTimeType | None = Field(
        sa_column=Column(sqlalchemy_utc.UtcDateTime, index=True)
    )
    endtime: CustomUTCDateTimeType | None = Field(
        sa_column=Column(sqlalchemy_utc.UtcDateTime, index=True)
    )
    network: str | None = None
    station: str | None = None
    channel: str | None = None
    location: str | None = None
    category: str | None = None
    priority: int | None = Field(default=1)
    data_valid: bool | None = Field(default=True, index=True)
    status: str | None = "new"
    # metadata field is already being used by sqlmodel and therefore needs a different name and alias
    metadata_: Dict[str, Any] | None = Field(
        sa_column=Column("metadata", JSON),
        alias="metadata",
        schema_extra={"serialization_alias": "metadata"},
    )
    comment: str | None = Field(sa_column=Column(Text))
    review_comment: str | None = Field(sa_column=Column(Text))

    __table_args__ = (
        Index(
            "index_station_metadata",
            # sncl
            "network",
            "station",
            "channel",
            "location",
            # type
            "category",
            # date
            "starttime",
            "endtime",
            # valid
            "data_valid",
            "status",
        ),
        Index(
            "index_category_time",
            # type
            "category",
            # date
            "starttime",
            "endtime",
        ),
    )

    # instructions for model_dump() to serialize aware datetime.datetime into aware datetime.datetime type
    # model_dump() serializes timestamps to strings. this returns the datetime.datetime type
    @field_serializer("created_time")
    def serialize_created_time(self, created_time: datetime):
        return created_time

    @field_serializer("updated_time")
    def serialize_updated_time(self, updated_time: datetime):
        return updated_time

    @field_serializer("starttime")
    def serialize_starttime(self, starttime: datetime):
        return starttime

    @field_serializer("endtime")
    def serialize_endtime(self, endtime: datetime):
        return endtime


# this pydantic model exists as an abstraction layer between the code and the database for 2 reasons
# 1. sqlmodel table validation is not powerful
# 2. this allows the metadata_ name conflict to only be an issue when interacting directly with the database in the factory
class Metadata(BaseModel):
    id: int | None = None
    created_by: str | None = None
    created_time: CustomUTCDateTimeType | None = None
    updated_by: str | None = None
    updated_time: CustomUTCDateTimeType | None = None
    starttime: CustomUTCDateTimeType | None = None
    endtime: CustomUTCDateTimeType | None = None
    network: str | None = None
    station: str | None = None
    channel: str | None = None
    location: str | None = None
    category: MetadataCategory | None = None
    priority: int | None = 1
    data_valid: bool | None = True
    status: str | None = "new"
    metadata: Dict[str, Any] = Field(default=None, alias="metadata_")
    comment: str | None = None
    review_comment: str | None = None

    # instructions for model_dump() to serialize pydantic CustomUTCDateTimeType into aware datetime.datetime type
    # sqlalchemy is expecting aware datetime.datetime, not the string model_dump() creates by default
    @field_serializer("created_time")
    def serialize_created_time(self, created_time: UTCDateTime):
        if created_time is not None:
            created_time = created_time.datetime.replace(tzinfo=timezone.utc)
        return created_time

    @field_serializer("updated_time")
    def serialize_updated_time(self, updated_time: UTCDateTime):
        if updated_time is not None:
            updated_time = updated_time.datetime.replace(tzinfo=timezone.utc)
        return updated_time

    @field_serializer("starttime")
    def serialize_starttime(self, starttime: UTCDateTime):
        if starttime is not None:
            starttime = starttime.datetime.replace(tzinfo=timezone.utc)
        return starttime

    @field_serializer("endtime")
    def serialize_endtime(self, endtime: UTCDateTime):
        if endtime is not None:
            endtime = endtime.datetime.replace(tzinfo=timezone.utc)
        return endtime


class MetadataQuery(BaseModel):
    id: int | None = None
    created_by: str | None = None
    category: MetadataCategory | None = None
    starttime: CustomUTCDateTimeType | None = None
    endtime: CustomUTCDateTimeType | None = None
    created_after: CustomUTCDateTimeType | None = None
    created_before: CustomUTCDateTimeType | None = None
    network: str | None = None
    station: str | None = None
    channel: str | None = None
    location: str | None = None
    data_valid: bool | None = None
    metadata: Dict[str, Any] | None = None
    status: List[str] | None = None


# this class validates the required id field
class MetadataUpdate(BaseModel):
    id: int
    created_by: str | None = None
    created_time: CustomUTCDateTimeType | None = None
    updated_by: str | None = None
    updated_time: CustomUTCDateTimeType | None = None
    starttime: CustomUTCDateTimeType | None = None
    endtime: CustomUTCDateTimeType | None = None
    network: str | None = None
    station: str | None = None
    channel: str | None = None
    location: str | None = None
    category: MetadataCategory | None = None
    priority: int | None = None
    data_valid: bool | None = None
    status: str | None = None
    metadata: Dict[str, Any] = Field(
        default=None,
        alias="metadata_",
        schema_extra={"serialization_alias": "metadata_"},
    )
    comment: str | None = None
    review_comment: str | None = None

    # instructions for model_dump() to serialize pydantic CustomUTCDateTimeType into aware datetime.datetime type
    # sqlalchemy is expecting aware datetime.datetime, not the string model_dump() creates by default
    @field_serializer("created_time")
    def serialize_created_time(self, created_time: UTCDateTime):
        if created_time is not None:
            created_time = created_time.datetime.replace(tzinfo=timezone.utc)
        return created_time

    @field_serializer("updated_time")
    def serialize_updated_time(self, updated_time: UTCDateTime):
        if updated_time is not None:
            updated_time = updated_time.datetime.replace(tzinfo=timezone.utc)
        return updated_time

    @field_serializer("starttime")
    def serialize_starttime(self, starttime: UTCDateTime):
        if starttime is not None:
            starttime = starttime.datetime.replace(tzinfo=timezone.utc)
        return starttime

    @field_serializer("endtime")
    def serialize_endtime(self, endtime: UTCDateTime):
        if endtime is not None:
            endtime = endtime.datetime.replace(tzinfo=timezone.utc)
        return endtime
