"""Module for metadata service.

Uses login.py for user management.

Anyone can access metadata.
Logged in users can create new metadata.
Update is restricted based on group membership.


Configuration:
    uses environment variables:

    REVIEWER_GROUP        - update is restricted the reviewer group.
"""

import os
from typing import List

from fastapi import APIRouter, Body, Depends, Response, Request, Query, HTTPException

from geomagio.api.db.models.metadata import (
    Metadata,
    MetadataUpdate,
    MetadataCategory,
    MetadataQuery,
)
from geomagio.api.db.models.metadata_history import MetadataHistory
from geomagio.api.db.factories.MetadataDatabaseFactory import MetadataDatabaseFactory
from geomagio.api.db.factories.MetadataHistoryDatabaseFactory import (
    MetadataHistoryDatabaseFactory,
)
from geomagio.api.db.database import engine
from .login_routes import require_user, User
from ...pydantic_utcdatetime import CustomUTCDateTimeType


router = APIRouter()


def get_metadata_query(
    category: MetadataCategory = None,
    created_by: str = None,
    starttime: CustomUTCDateTimeType = None,
    endtime: CustomUTCDateTimeType = None,
    created_after: CustomUTCDateTimeType = None,
    created_before: CustomUTCDateTimeType = None,
    network: str = None,
    station: str = None,
    channel: str = None,
    location: str = None,
    data_valid: bool = None,
    status: List[str] = Query(None),
) -> MetadataQuery:
    return MetadataQuery(
        category=category,
        created_by=created_by,
        starttime=starttime,
        endtime=endtime,
        created_after=created_after,
        created_before=created_before,
        network=network,
        station=station,
        channel=channel,
        location=location,
        data_valid=data_valid,
        status=status,
    )


@router.post(
    "/metadata",
    description="Save metadata in database",
    response_model=Metadata,
    status_code=201,
)
async def create_metadata(
    request: Request,
    metadata: Metadata,
    user: User = Depends(require_user()),
):
    return await MetadataDatabaseFactory(engine=engine).create_metadata(
        new_metadata=metadata
    )


@router.get(
    "/metadata",
    description="Search metadata by query parameters",
    name="Request metadata",
    response_model=List[Metadata],
)
async def get_metadata(
    # each field needs to be handled separately otherwise FastAPI thinks it should expect a body
    query: MetadataQuery = Depends(get_metadata_query),
):
    return await MetadataDatabaseFactory(engine=engine).get_metadata(query=query)


@router.get(
    "/metadata/history",
    description="Search historical metadata by query parameters",
    response_model=List[MetadataHistory],
)
async def get_metadata_history(
    # each field needs to be handled separately otherwise FastAPI thinks it should expect a body
    query: MetadataQuery = Depends(get_metadata_query),
):
    return await MetadataHistoryDatabaseFactory(engine=engine).get_metadata_history(
        query=query
    )


@router.get(
    "/metadata/{id}",
    description="Search metadata by database id",
    response_model=Metadata,
)
async def get_metadata_by_id(id: int):
    metadata = await MetadataDatabaseFactory(engine=engine).get_metadata_by_id(id=id)
    if metadata is None:
        raise HTTPException(status_code=404, detail="metadata not found")
    else:
        return metadata


@router.get(
    "/metadata/{metadata_id}/history",
    description="Search metadata version history by database id",
)
async def get_metadata_history_by_metadata_id(
    metadata_id: int,
):
    metadata_history = await MetadataHistoryDatabaseFactory(
        engine=engine
    ).get_metadata_history_by_metadata_id(
        metadata_id=metadata_id,
    )
    if metadata_history is None:
        raise HTTPException(status_code=404, detail="metadata not found")
    else:
        return metadata_history


@router.get(
    "/metadata/history/{id}",
    description="Search historical metadata by database id",
    response_model=MetadataHistory,
)
async def get_metadata_history_by_id(id: int):
    metadata_history = await MetadataHistoryDatabaseFactory(
        engine=engine
    ).get_metadata_history_by_id(id=id)
    if metadata_history is None:
        raise HTTPException(status_code=404, detail="metadata_history not found")
    return metadata_history


@router.put(
    "/metadata/{id}",
    description="Edit metadata from older version",
    response_model=Metadata,
)
async def update_metadata(
    id: int,
    metadata: MetadataUpdate,
    user: User = Depends(
        require_user(
            [os.getenv("REVIEWER_GROUP", "ghsc/geomag/operations/roles/reviewer")]
        )
    ),
):
    try:
        return await MetadataDatabaseFactory(engine=engine).update_metadata(
            updated_metadata=metadata,
            updated_by=user.nickname,
        )
    except Exception as e:
        if str(e) == "metadata not found":
            raise HTTPException(status_code=404, detail="metadata not found")
