import json

from fastapi import APIRouter, Depends, HTTPException, Query
from starlette.responses import Response


from geomagio.algorithm.DbDtAlgorithm import DbDtAlgorithm
from geomagio.residual.Calculation import calculate
from geomagio.residual.Reading import Reading

from .DataApiQuery import DataApiQuery
from .FilterApiQuery import FilterApiQuery
from .data import format_timeseries, get_data_factory, get_data_query, get_timeseries
from .filter import get_filter_data_query
from . import filter
import logging

logger = logging.getLogger(__name__)

router = APIRouter()


@router.get(
    "/algorithms/dbdt/",
    description="First order derivative at requested interval",
    name="Dbdt Algorithm",
)
def get_dbdt(
    query: DataApiQuery = Depends(get_data_query),
) -> Response:
    dbdt = DbDtAlgorithm(period=query.sampling_period)
    data_factory = get_data_factory(query=query)
    # read data
    raw = get_timeseries(data_factory, query)
    # run dbdt
    timeseries = dbdt.process(raw)
    elements = [f"{element}_DT" for element in query.elements]
    # output response
    return format_timeseries(
        timeseries=timeseries, format=query.format, elements=elements
    )


####################################### The router .get filter isnt visible on the docs page
# Look for register routers in the backend
@router.get(
    "/algorithms/filter/",
    description="Filtered data dependent on requested interval",
    name="Filtered Algorithm",
)
def get_filter(query: FilterApiQuery = Depends(get_filter_data_query)) -> Response:

    filtered_timeseries = filter.get_timeseries(query)

    elements = [f"{element}" for element in query.elements]
    return format_timeseries(
        timeseries=filtered_timeseries, format=query.format, elements=elements
    )


###############################################


@router.post(
    "/algorithms/residual",
    description="Caclulates new absolutes and baselines from reading\n\n"
    + "Returns posted reading object with updated values",
    response_model=Reading,
)
def calculate_residual(reading: Reading, adjust_reference: bool = True):
    try:
        calculated = calculate(
            reading=reading, adjust_reference=adjust_reference
        ).json()
        return json.loads(
            # replace non-JSON compliant float values with null string
            calculated.replace("NaN", "null")
            .replace("-Infinity", "null")  # must replace "-Infinity" before "Infinity"
            .replace("Infinity", "null")
        )
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))
