from typing import List, Union, Optional
from fastapi import Query
from obspy import Stream
import logging as logger

import numpy as np

from geomagio import TimeseriesUtility
from .DataApiQuery import (
    DEFAULT_ELEMENTS,
    DataHost,
    DataType,
    OutputFormat,
    SamplingPeriod,
)
from .FilterApiQuery import FilterApiQuery
from geomagio.algorithm.FilterAlgorithm import FilterAlgorithm
from .data import get_data_factory
from geomagio.pydantic_utcdatetime import CustomUTCDateTimeType


def get_filter_data_query(
    id: str = Query(..., title="Observatory code"),
    starttime: Optional[CustomUTCDateTimeType] = Query(
        None,
        title="Start Time",
        description="Time of first requested data. Default is start of current UTC day.",
    ),
    endtime: Optional[CustomUTCDateTimeType] = Query(None, title="End Time"),
    elements: List[str] = Query(DEFAULT_ELEMENTS, title="Geomagnetic Elements"),
    data_type: Union[DataType, str] = Query(DataType.VARIATION, alias="type"),
    format: Union[OutputFormat, str] = Query(
        OutputFormat.IAGA2002, title="Output Format"
    ),
    input_sampling_period: Optional[SamplingPeriod] = Query(
        None,
        title="Input Sampling Period",
        description="`--` dynamically determines a necessary sampling period.",
    ),
    sampling_period: Optional[SamplingPeriod] = Query(
        None,
        alias="output_sampling_period",
        title="Output sampling period",
    ),
    data_host: Union[DataHost, str] = Query(
        DataHost.DEFAULT, title="Data Host", description="Edge host to pull data from."
    ),
) -> FilterApiQuery:

    return FilterApiQuery(
        id=id,
        starttime=starttime,
        endtime=endtime,
        elements=elements,
        input_sampling_period=input_sampling_period,
        sampling_period=sampling_period,
        data_type=data_type,
        data_host=data_host,
        format=format,
    )


# Main filter function
def get_timeseries(query: FilterApiQuery) -> Stream:
    data_factory = get_data_factory(query=query)

    # Determine input sampling period if not provided
    if query.input_sampling_period is None:
        # Dynamically determine the input sampling period
        input_sampling_period, data = determine_available_period(
            query.sampling_period, query, data_factory
        )
    else:
        input_sampling_period = query.input_sampling_period

    filt = FilterAlgorithm(
        input_sample_period=input_sampling_period,
        output_sample_period=query.sampling_period,
    )
    # Fetch filtered data
    starttime, endtime = filt.get_input_interval(query.starttime, query.endtime)

    data = data_factory.get_timeseries(
        starttime=starttime,
        endtime=endtime,
        observatory=query.id,
        channels=query.elements,
        type=query.data_type,
        interval=TimeseriesUtility.get_interval_from_delta(filt.input_sample_period),
    )

    # Apply filtering if needed

    filtered_timeseries = filt.process(data)
    return filtered_timeseries


def determine_available_period(output_sampling_period: float, query, data_factory):
    """
    Finds the lowest resolution (longest sampling period) <= output_sampling_period
    that has valid data available.
    """

    # Sort and filter periods starting from output_sampling_period
    sorted_periods: List[SamplingPeriod] = sorted(
        SamplingPeriod, key=lambda p: p.value, reverse=True
    )
    if output_sampling_period is None:
        raise ValueError("Output sampling period cannot be None.")
    else:

        valid_sampling_periods = [
            p for p in sorted_periods if p.value <= output_sampling_period
        ]

    for period in valid_sampling_periods:
        if period <= output_sampling_period:

            data = data_factory.get_timeseries(
                starttime=query.starttime,
                endtime=query.endtime,
                observatory=query.id,
                channels=query.elements,
                type=query.data_type,
                interval=TimeseriesUtility.get_interval_from_delta(period.value),
            )
            # Check if the fetched data is valid
            if is_valid_data(data):

                logger.info(f"Valid data found for sampling period: {period.name}")
                return period.value, data  # Return the sampling period and the data

        else:
            logger.error(
                f"No valid data found for requested sampling period: {period.name}"
            )
            continue

    raise ValueError("No valid data found for the requested output sampling period.")


def is_valid_data(data: Stream) -> bool:
    """
    Checks if the fetched data contains actual values and not just filler values (e.g., NaN).
    A Stream is invalid if any trace contains only NaN values.
    """
    if not data or len(data) == 0:
        return False  # No data in the stream

    for trace in data:
        # Check if trace.data exists and has data
        if trace.data is None or len(trace.data) == 0:
            return False  # Trace has no data

        # Check if all values in trace.data are NaN
        if np.all(np.isnan(trace.data)):
            return False  # Invalid if all values are NaN

    return True  # All traces are valid
