import json
import numpy as np
from obspy import Stream, Trace, UTCDateTime
from geomagio import ChannelConverter
from geomagio.TimeseriesFactoryException import TimeseriesFactoryException


class IMFJSONParser(object):

    @staticmethod
    def parse(data):
        try:
            json_data = json.loads(data)
            # Extract global metadata and map to standard keys.
            global_meta = json_data.get("metadata", {})
            intermag = global_meta.get("intermagnet", {})
            imo = intermag.get("imo", {})

            # Map IMO fields to ObsPy Stats defaults.
            default_station = imo.get("iaga_code")
            default_station_name = imo.get("name")
            coordinates = imo.get("coordinates", [])
            default_longitude = coordinates[0] if len(coordinates) >= 1 else None
            default_latitude = coordinates[1] if len(coordinates) >= 2 else None
            default_elevation = coordinates[2] if len(coordinates) >= 3 else None

            # Extra metadata from global metadata.
            default_data_type = intermag.get("data_type")
            default_sampling_period = intermag.get("sampling_period")
            default_digital_sampling_rate = intermag.get("digital_sampling_rate", 0)
            default_sensor_orientation = intermag.get("sensor_orientation")
            default_reported_orientation = intermag.get("reported_orientation")
            default_source = global_meta.get("source_of_data")

            # Process common time stamps.
            times = json_data.get("times")
            if not times or not isinstance(times, list):
                raise TimeseriesFactoryException(
                    "Missing or invalid 'times' in JSON response."
                )
            try:
                starttime = UTCDateTime(times[0])
                if len(times) >= 2:
                    delta = UTCDateTime(times[1]) - UTCDateTime(times[0])
                elif default_sampling_period is not None:
                    delta = float(default_sampling_period)
                else:
                    raise Exception("Unable to determine time delta.")
            except Exception as e:
                raise TimeseriesFactoryException("Error processing times: " + str(e))

            stream = Stream()
            values_list = json_data.get("values")
            if not values_list or not isinstance(values_list, list):
                raise TimeseriesFactoryException(
                    "Missing or invalid 'values' in JSON response."
                )

            for entry in values_list:
                entry_meta = entry.get("metadata", {})
                # Use 'channel' (or fall back to 'element') from the entry.
                channel = entry_meta.get("channel") or entry_meta.get("element")
                if not channel:
                    continue

                # Map entry metadata, using global defaults when necessary.
                trace_station = entry_meta.get("station") or default_station
                trace_network = entry_meta.get("network") or "NT"
                trace_location = entry_meta.get("location") or ""

                # Build the ObsPy Stats dictionary.
                stats = {}
                stats["station"] = trace_station
                stats["network"] = trace_network
                stats["location"] = trace_location
                stats["channel"] = channel
                stats["starttime"] = starttime
                stats["delta"] = float(delta)
                # Extra metadata using conventional names.
                if default_station_name:
                    stats["station_name"] = default_station_name
                if default_longitude is not None:
                    stats["geodetic_longitude"] = float(default_longitude)
                if default_latitude is not None:
                    stats["geodetic_latitude"] = float(default_latitude)
                if default_elevation is not None:
                    stats["elevation"] = float(default_elevation)
                stats["data_type"] = default_data_type
                stats["sampling_period"] = default_sampling_period
                if default_digital_sampling_rate > 0:
                    stats["sensor_sampling_rate"] = (
                        1 / default_digital_sampling_rate
                    )  # because webservice returns 1/x, so flip to return to original.
                stats["sensor_orientation"] = default_sensor_orientation
                stats["reported_orientation"] = default_reported_orientation
                stats["source_of_data"] = default_source
                # Derive interval type.
                sp = default_sampling_period
                if sp == 60:
                    stats["data_interval_type"] = "1-minute"
                elif sp == 1:
                    stats["data_interval_type"] = "1-second"
                elif sp == 3600:
                    stats["data_interval_type"] = "1-hour"
                elif sp == 86400:
                    stats["data_interval_type"] = "1-day"
                else:
                    stats["data_interval_type"] = f"{sp}-second"

                trace_values = entry.get("values")
                if trace_values is None:
                    continue

                trace = Trace(data=np.array(trace_values, dtype=float), header=stats)
                # convert from arcmin (returned from geomag ws using imfjson format)
                if "D" == channel:
                    trace.data = ChannelConverter.get_radians_from_minutes(trace.data)
                # Attach metadata to each trace.
                trace.stats.metadata = global_meta
                stream.append(trace)
            return stream
        except Exception as e:
            raise TimeseriesFactoryException("Error Processing JSON: " + str(e))
