import os
import requests
import datetime
from typing import List, Union
import json

from obspy import UTCDateTime
from pydantic import TypeAdapter

from geomagio.api.db.models.metadata import Metadata, MetadataQuery

# Force IPV4 in requests package
requests.packages.urllib3.util.connection.HAS_IPV6 = False

# Set the API host and URL for the Geomagnetic Metadata service
GEOMAG_API_HOST = os.getenv("GEOMAG_API_HOST", "geomag.usgs.gov")
GEOMAG_API_URL = f"https://{GEOMAG_API_HOST}/ws/secure/metadata"
if "127.0.0.1" in GEOMAG_API_URL:
    GEOMAG_API_URL = GEOMAG_API_URL.replace("https://", "http://")


class MetadataFactory(object):
    """
    A factory class for interacting with metadata servers.

    Attributes:
    ----------
        url: str
            The base URL for the metadata server.
        token: str
            The authorization token for accessing the API.
    """

    def __init__(
        self,
        url: str = GEOMAG_API_URL,
        token: str = os.getenv("GITLAB_API_TOKEN"),
    ):
        """
        Initializes the MetadataFactory with the specified URL and token.

        Parameters:
        -----------
            url: str
                The base URL for the metadata server (default is GEOMAG_API_URL).
            token: str
                The authorization token (default is fetched from environment).
        """
        self.url = url
        self.token = token

    def _get_headers(self):
        """
        Constructs the headers for the API request.

        Returns:
        --------
            dict:
                A dictionary containing the authorization token and content type.
        """
        return (
            {"Authorization": self.token, "content-type": "application/json"}
            if self.token
            else None
        )

    def get_metadata(
        self, query: MetadataQuery, timeout: Union[float, tuple] = None
    ) -> List[Metadata]:
        """
        Retrieves metadata based on the provided query.

        Parameters:
        -----------
            query: MetadataQuery
                The query object containing search parameters.
            timeout: float or tuple, optional
                The timeout for the API request.
                Can be a single number (for both connection and read timeouts)
                or a tuple (connection timeout, read timeout). Defaults to None.

        Returns:
        --------
            metadata: List[Metadata]
                A list of Metadata objects matching the query.
        """
        if query.id:
            metadata = [self.get_metadata_by_id(id=query.id)]
        else:
            response = requests.get(
                url=self.url,
                headers=self._get_headers(),
                params=parse_params(query=query),
                timeout=timeout,
            )
            metadataAdapter = TypeAdapter(List[Metadata])
            metadata = metadataAdapter.validate_python(response.json())

            # Filter metadata based on additional criteria if provided
            if query.metadata is not None:
                query_metadata_items = query.metadata.items()
                metadata_list = metadata

                metadata = [
                    metadata
                    for metadata in metadata_list
                    if all(
                        metadata.metadata.get(key) == value
                        for key, value in query_metadata_items
                    )
                ]

        return metadata

    def get_metadata_by_id(
        self, id: int, timeout: Union[float, tuple] = None
    ) -> Metadata:
        """
        Retrieves metadata for a specific ID.

        Parameters:
        -----------
            id: int
                The ID of the metadata to retrieve.
            timeout: float or tuple, optional
                The timeout for the API request.
                Can be a single number (for both connection and read timeouts)
                or a tuple (connection timeout, read timeout). Defaults to None.

        Returns:
        --------
            Metadata: object
                The Metadata object corresponding to the given ID.
        """
        response = requests.get(
            url=f"{self.url}/{id}", headers=self._get_headers(), timeout=timeout
        )
        return Metadata(**response.json())

    def create_metadata(
        self, metadata: Metadata, timeout: Union[float, tuple] = None
    ) -> Metadata:
        """
        Creates new metadata objects in the database.

        Parameters:
        -----------
            metadata: Metadata
                The Metadata object to create.
            timeout: float or tuple, optional
                The timeout for the API request.
                Can be a floating-point number (for both connection and read timeouts)
                or a tuple (connection timeout, read timeout). Defaults to None.

        Returns:
        --------
            Metadata: The created Metadata object.
        """
        response = requests.post(
            url=self.url,
            data=metadata.model_dump_json(),
            headers=self._get_headers(),
            timeout=timeout,
        )
        return Metadata(**response.json())

    def update_metadata(
        self, metadata: Metadata, timeout: Union[float, tuple] = None
    ) -> Metadata:
        """
        Updates existing metadata in the system.

        Parameters:
        -----------
            metadata: Metadata
                The Metadata object to update.
            timeout: float or tuple, optional
                The timeout for the API request.
                Can be a floating-point number (for both connection and read timeouts)
                or a tuple (connection timeout, read timeout). Defaults to None.

        Returns:
        -----------
            Metadata:
                The updated Metadata object.
        """
        response = requests.put(
            url=f"{self.url}/{metadata.id}",
            data=metadata.model_dump_json(),
            headers=self._get_headers(),
            timeout=timeout,
        )
        return Metadata(**response.json())


def parse_params(query: MetadataQuery) -> str:
    """
    Parses the query object into a dictionary of parameters for the API request.

    Parameters:
    -----------
        query: MetadataQuery
            The query object containing search parameters.

    Returns:
    --------
        dict:
            A dictionary of parameters formatted for the API request.
    """
    query = query.model_dump(exclude_none=True)
    args = {}
    for key in query.keys():
        element = query[key]
        # Convert times to ISO format strings
        if isinstance(element, UTCDateTime) or isinstance(element, datetime.datetime):
            element = element.isoformat()
        # Serialize the metadata dictionary to a JSON string
        elif key == "metadata" and isinstance(element, dict):
            args[key] = json.dumps(element)
        # Get string value of metadata category
        if key == "category":
            element = element.value
        args[key] = element
    return args
