from enum import Enum
from typing import List, Optional

from typer import Argument, Option, Typer

from geomagio.algorithm.Algorithm import Algorithm
from geomagio.algorithm.FilterAlgorithm import FilterAlgorithm
from geomagio.Controller import Controller, get_realtime_interval
from geomagio.geomag_types import DataInterval
from geomagio.TimeseriesFactory import TimeseriesFactory
from geomagio.edge.EdgeFactory import EdgeFactory
from geomagio.edge.MiniSeedFactory import MiniSeedFactory


class DataFormat(str, Enum):
    OBSRIO = "OBSRIO"
    PCDCP = "PCDCP"


app = Typer(help="Filter geomagnetic timeseries data")


def main():
    app()


@app.command(
    name="day",
    help="Filter 1 day nT/temperature data",
)
def day_command(
    observatory: str = Argument(None, help="observatory id"),
    input_host: str = Option("127.0.0.1", help="host to request data from"),
    output_host: str = Option("127.0.0.1", help="host to write data to"),
    input_port: int = Option(None, help="port to pull inputs from on input host"),
    output_read_port: int = Option(None, help="port to check for gaps on output host"),
    output_port: int = Option(None, help="port to write to on output host"),
    realtime_interval: int = Option(
        604800, help="length of update window (in seconds)"
    ),
    update_limit: int = Option(4, help="number of update windows"),
):
    day_filter(
        observatory=observatory,
        channels=("U", "V", "W", "F", "T1", "T2", "T3", "T4"),
        input_factory=EdgeFactory(
            host=input_host,
            port=input_port,
            type="variation",
            sncl_mode="geomag",
        ),
        output_factory=EdgeFactory(
            host=output_host,
            port=output_read_port,
            write_port=output_port,
            type="variation",
            sncl_mode="geomag",
        ),
        realtime_interval=realtime_interval,
        update_limit=update_limit,
    )
    # remove the following after data migration is complete
    _copy_channels(
        observatory=observatory,
        channels=(
            ("U", "H"),
            ("V", "E"),
            ("W", "Z"),
            ("F", "F"),
        ),
        interval="day",
        input_factory=EdgeFactory(
            host=input_host,
            port=input_port,
            type="variation",
            sncl_mode="geomag",
        ),
        output_factory=EdgeFactory(
            host=output_host,
            port=output_read_port,
            write_port=7981,  # hard-coded rawinput until migration, then remove
            type="variation",
            sncl_mode="legacy",
        ),
        realtime_interval=realtime_interval,
        update_limit=update_limit,
    )


@app.command(
    name="hour",
    help="Filter 1 hour nT/temperature data",
)
def hour_command(
    observatory: str = Argument(None, help="observatory id"),
    input_host: str = Option("127.0.0.1", help="host to request data from"),
    output_host: str = Option("127.0.0.1", help="host to write data to"),
    input_port: int = Option(None, help="port to pull inputs from on input host"),
    output_read_port: int = Option(None, help="port to check for gaps on output host"),
    output_port: int = Option(None, help="port to write to on output host"),
    realtime_interval: int = Option(86400, help="length of update window (in seconds)"),
    update_limit: int = Option(7, help="number of update windows"),
):
    hour_filter(
        observatory=observatory,
        channels=("U", "V", "W", "F", "T1", "T2", "T3", "T4"),
        input_factory=EdgeFactory(
            host=input_host,
            port=input_port,
            type="variation",
            sncl_mode="geomag",
        ),
        output_factory=EdgeFactory(
            host=output_host,
            port=output_read_port,
            write_port=output_port,
            type="variation",
            sncl_mode="geomag",
        ),
        realtime_interval=realtime_interval,
        update_limit=update_limit,
    )
    # remove the following after data migration is complete
    _copy_channels(
        observatory=observatory,
        channels=(
            ("U", "H"),
            ("V", "E"),
            ("W", "Z"),
            ("F", "F"),
        ),
        interval="hour",
        input_factory=EdgeFactory(
            host=input_host,
            port=input_port,
            type="variation",
            sncl_mode="geomag",
        ),
        output_factory=EdgeFactory(
            host=output_host,
            port=output_read_port,
            write_port=7981,  # hard-coded rawinput until migration, then remove
            type="variation",
            sncl_mode="legacy",
        ),
        realtime_interval=realtime_interval,
        update_limit=update_limit,
    )


@app.command(
    name="realtime",
    short_help="Filter 1 second and 1 minute nT/temperature data",
    help="""
    ObsRIO:

        Filters 10Hz U,V,W to 1 second U,V,W

        Filters 1 second U,V,W,F to 1 minute U,V,W,F

        Filters 1 second T1-4 to 1 minute UK1-4

        Copies 1 second and 1 minute U,V,W,F miniseed to legacy
        H,E,Z,F 

    PCDCP:

        Copies 1 second H,E,Z,F earthworm to U,V,W,F miniseed

        Filters 1 second U,V,W,F miniseed to 1 minute miniseed

    """,
)
def realtime_command(
    observatory: str = Argument(None, help="observatory id"),
    input_host: str = Option("127.0.0.1", help="host to request data from"),
    output_host: str = Option("127.0.0.1", help="host to write data to"),
    input_port: int = Option(None, help="port to pull inputs from on input host"),
    output_read_port: int = Option(None, help="port to check for gaps on output host"),
    output_port: int = Option(None, help="port to write to on output host"),
    data_format: DataFormat = Option(DataFormat.PCDCP, help="Data acquisition system"),
    realtime_interval: int = Option(600, help="length of update window (in seconds)"),
    update_limit: int = Option(10, help="number of update windows"),
):
    if data_format == DataFormat.OBSRIO:
        second_filter(
            observatory=observatory,
            input_factory=MiniSeedFactory(
                host=input_host,
                port=None,  # use MiniSeedFactory default
                convert_channels=("U", "V", "W"),
                type="variation",
                sncl_mode="geomag",
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,
                write_port=output_port,
                type="variation",
                sncl_mode="geomag",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )
        # remove the following after data migration is complete
        _copy_channels(
            observatory=observatory,
            channels=(
                ("U", "H"),
                ("V", "E"),
                ("W", "Z"),
            ),
            interval="second",
            input_factory=EdgeFactory(
                host=input_host, port=input_port, type="variation", sncl_mode="geomag"
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,
                write_port=7981,  # hard-code port for legacy sncls
                type="variation",
                sncl_mode="legacy",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )
        # remove the following after data migration is complete
        _copy_channels(
            observatory=observatory,
            channels=(("F", "F"),),
            interval="second",
            input_factory=MiniSeedFactory(
                host=input_host, port=input_port, type="variation", sncl_mode="geomag"
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,
                write_port=7981,  # hard-code port for legacy sncls
                type="variation",
                sncl_mode="legacy",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )
        minute_filter(
            observatory=observatory,
            channels=("H", "E", "Z"),
            input_factory=EdgeFactory(
                host=input_host,
                port=input_port,  # earthworm port required for realtime
                type="variation",
                sncl_mode="geomag",
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,  # earthworm port required for realtime
                write_port=output_port,  # rawinput port required for realtime
                type="variation",
                sncl_mode="geomag",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )
        # segregate F and T* filtering from U, V, and W until/unless ObsRio
        # transmits scaled integer data
        minute_filter(
            observatory=observatory,
            channels=("F", "T1", "T2", "T3", "T4", "T5"),
            input_factory=MiniSeedFactory(
                host=input_host,
                port=None,  # use MiniSeedFactory default
                type="variation",
                sncl_mode="geomag",
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,  # earthworm port required for realtime
                write_port=output_port,  # rawinput port required for realtime
                type="variation",
                sncl_mode="geomag",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )
        # remove the following after data migration is complete
        _copy_channels(
            observatory=observatory,
            channels=(
                ("H", "H"),
                ("E", "E"),
                ("Z", "Z"),
                ("F", "F"),
            ),
            interval="minute",
            input_factory=EdgeFactory(
                host=input_host,
                port=input_port,
                type="variation",
                sncl_mode="geomag",
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,
                write_port=7981,  # hard-coded rawinput until migration, then remove
                type="variation",
                sncl_mode="legacy",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )

    else:
        # legacy PCDCP processing
        _copy_channels(
            observatory=observatory,
            channels=(
                ("H", "H"),
                ("E", "E"),
                ("Z", "Z"),
                ("F", "F"),
            ),
            interval="second",
            input_factory=EdgeFactory(
                host=input_host,
                port=input_port,
                type="variation",
                sncl_mode="legacy",
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,
                write_port=output_port,
                type="variation",
                sncl_mode="geomag",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )
        # legacy PCDCP processing
        _copy_channels(
            observatory=observatory,
            channels=(
                ("H", "H"),
                ("E", "E"),
                ("Z", "Z"),
                ("F", "F"),
            ),
            interval="minute",
            input_factory=EdgeFactory(
                host=input_host,
                port=input_port,
                type="variation",
                sncl_mode="legacy",
            ),
            output_factory=EdgeFactory(
                host=output_host,
                port=output_read_port,
                write_port=output_port,
                type="variation",
                sncl_mode="geomag",
            ),
            realtime_interval=realtime_interval,
            update_limit=update_limit,
        )


def day_filter(
    observatory: str,
    channels: List[str],
    input_factory: TimeseriesFactory,
    output_factory: TimeseriesFactory,
    realtime_interval: int,
    update_limit: int,
):
    """Filter 1 second miniseed channels to 1 day

    Parameters:
    -----------
    observatory: str
        observatory id
    channels: array
        list of channels to filter
    input_factory: TimeseriesFactory
        factory to request data
    output_factory: TimeseriesFactory
        factory to write data
    realtime_interval: int
        length of update window (in seconds)
    update_limit: int
        number of update windows
    """
    starttime, endtime = get_realtime_interval(realtime_interval, "day")
    controller = Controller(
        inputFactory=input_factory,
        inputInterval="minute",
        outputFactory=output_factory,
        outputInterval="day",
    )
    for channel in channels:
        controller.run_as_update(
            algorithm=FilterAlgorithm(
                input_sample_period=60.0,
                output_sample_period=86400.0,
                inchannels=(channel,),
                outchannels=(channel,),
            ),
            observatory=(observatory,),
            output_observatory=(observatory,),
            starttime=starttime,
            endtime=endtime,
            input_channels=(channel,),
            output_channels=(channel,),
            realtime=realtime_interval,
            update_limit=update_limit,
        )


def hour_filter(
    observatory: str,
    channels: List[str],
    input_factory: TimeseriesFactory,
    output_factory: TimeseriesFactory,
    realtime_interval: int,
    update_limit: int,
):
    """Filter 1 minute miniseed channels to 1 hour

    Parameters:
    -----------
    observatory: str
        observatory id
    channels: array
        list of channels to filter
    input_factory: TimeseriesFactory
        factory to request data
    output_factory: TimeseriesFactory
        factory to write data
    realtime_interval: int
        length of update window (in seconds)
    update_limit: int
        number of update windows
    """
    starttime, endtime = get_realtime_interval(realtime_interval, "hour")
    controller = Controller(
        inputFactory=input_factory,
        inputInterval="minute",
        outputFactory=output_factory,
        outputInterval="hour",
    )
    for channel in channels:
        controller.run_as_update(
            algorithm=FilterAlgorithm(
                input_sample_period=60.0,
                output_sample_period=3600.0,
                inchannels=(channel,),
                outchannels=(channel,),
            ),
            observatory=(observatory,),
            output_observatory=(observatory,),
            starttime=starttime,
            endtime=endtime,
            input_channels=(channel,),
            output_channels=(channel,),
            realtime=realtime_interval,
            update_limit=update_limit,
        )


def minute_filter(
    observatory: str,
    channels: List[str],
    input_factory: TimeseriesFactory,
    output_factory: TimeseriesFactory,
    realtime_interval: int = 600,
    update_limit: int = 10,
):
    """Filter 1 second miniseed channels to 1 minute

    Parameters:
    -----------
    observatory: str
        observatory id
    channels: array
        list of channels to filter
    type: str
        data type (e.g., "variation", "adjusted", or "R0", "A0")
    input_factory: TimeseriesFactory
        factory to request data
    output_factory: TimeseriesFactory
        factory to write data
    realtime_interval: int
        length of update window (in seconds)
    update_limit: int
        number of update windows
    """
    starttime, endtime = get_realtime_interval(realtime_interval, "minute")
    controller = Controller(
        inputFactory=input_factory,
        inputInterval="second",
        outputFactory=output_factory,
        outputInterval="minute",
    )
    for channel in channels:
        controller.run_as_update(
            algorithm=FilterAlgorithm(
                input_sample_period=1,
                output_sample_period=60,
                inchannels=(channel,),
                outchannels=(channel,),
            ),
            observatory=(observatory,),
            output_observatory=(observatory,),
            starttime=starttime,
            endtime=endtime,
            input_channels=(channel,),
            output_channels=(channel,),
            realtime=realtime_interval,
            update_limit=update_limit,
        )


def second_filter(
    observatory: str,
    channels: List[str] = ["U", "V", "W"],
    input_factory: Optional[TimeseriesFactory] = None,
    output_factory: Optional[TimeseriesFactory] = None,
    realtime_interval: int = 600,
    update_limit: int = 10,
):
    """Filter 10Hz miniseed U,V,W to 1 second

    Parameters:
    -----------
    observatory: str
        observatory id
    channels: array
        list of channels to filter
    input_factory: TimeseriesFactory
        factory to request data
    output_factory: TimeseriesFactory
        factory to write data
    realtime_interval: int
        length of update window (in seconds)
    update_limit: int
        number of update windows
    """
    starttime, endtime = get_realtime_interval(realtime_interval, "second")
    controller = Controller(
        inputFactory=input_factory,
        inputInterval="tenhertz",
        outputFactory=output_factory,
        outputInterval="second",
    )
    for channel in channels:
        controller.run_as_update(
            algorithm=FilterAlgorithm(
                input_sample_period=0.1,
                output_sample_period=1,
                inchannels=(channel,),
                outchannels=(channel,),
            ),
            observatory=(observatory,),
            output_observatory=(observatory,),
            starttime=starttime,
            endtime=endtime,
            input_channels=(channel,),
            output_channels=(channel,),
            realtime=realtime_interval,
            update_limit=update_limit,
        )


def temperature_filter(
    observatory: str,
    channels: List[List[str]],
    input_factory: TimeseriesFactory,
    input_interval: int,
    output_factory: TimeseriesFactory,
    output_interval: int,
    realtime_interval: int,
    update_limit: int,
):
    """Filter temperatures 1Hz miniseed (LK1-4) to 1 minute legacy (UK1-4)."""
    starttime, endtime = get_realtime_interval(realtime_interval, output_interval)
    controller = Controller(
        inputFactory=input_factory,
        inputInterval=input_interval,
        outputFactory=output_factory,
        outputInterval=output_interval,
    )
    for input_channel, output_channel in channels:
        controller.run_as_update(
            algorithm=FilterAlgorithm(
                input_sample_period=input_interval,
                output_sample_period=output_interval,
                inchannels=(input_channel,),
                outchannels=(output_channel,),
            ),
            observatory=(observatory,),
            output_observatory=(observatory,),
            starttime=starttime,
            endtime=endtime,
            input_channels=(input_channel,),
            output_channels=(output_channel,),
            realtime=realtime_interval,
            rename_output_channel=((input_channel, output_channel),),
            update_limit=update_limit,
        )


def _copy_channels(
    observatory: str,
    channels: List[List[str]],
    interval: DataInterval,
    input_factory: Optional[TimeseriesFactory],
    output_factory: Optional[TimeseriesFactory],
    realtime_interval: int = 600,
    update_limit: int = 10,
):
    """copy channels between earthworm and miniseed formats

    Parameters:
    -----------
    observatory: str
        observatory id
    channels: array
        list of channel conversions
        format: ((input_channel_1, output_channel_1), ...)
    interval: {tenhertz, second, minute, hour, day}
        data interval
    input_factory: TimeseriesFactory
        factory to request data
    output_factory: TimeseriesFactory
        factory to write data
    realtime_interval: int
        length of update window (in seconds)
    update_limit: int
        number of update windows
    """
    starttime, endtime = get_realtime_interval(realtime_interval, interval)
    controller = Controller(
        inputFactory=input_factory,
        inputInterval=interval,
        outputFactory=output_factory,
        outputInterval=interval,
    )
    for input_channel, output_channel in channels:
        controller.run_as_update(
            algorithm=Algorithm(
                inchannels=(input_channel,),
                outchannels=(output_channel,),
            ),
            observatory=(observatory,),
            output_observatory=(observatory,),
            starttime=starttime,
            endtime=endtime,
            input_channels=(input_channel,),
            output_channels=(output_channel,),
            rename_output_channel=((input_channel, output_channel),),
            realtime=realtime_interval,
            update_limit=update_limit,
        )
