import json
import os
import numpy as np
from obspy import Trace, Stream, UTCDateTime

from geomagio.ObservatoryMetadata import ObservatoryMetadata
from geomagio.TimeseriesFactory import TimeseriesFactory
from geomagio.TimeseriesFactoryException import TimeseriesFactoryException

from io import BytesIO
from typing import List, Literal, Optional, Union, cast
import json
import numpy
from obspy import Stream, Trace, UTCDateTime
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from geomagio.api.ws.DataApiQuery import DataApiQuery
from geomagio.api.ws.FilterApiQuery import FilterApiQuery
from geomagio.geomag_types import DataType
from geomagio.imfjson.IMFJSONParser import IMFJSONParser


class WSFactory(TimeseriesFactory):
    """
    WSFactory retrieves timeseries data from a web service URL.
    This is an input-only factory.
    """

    def __init__(
        self,
        observatory: Optional[str] = None,
        channels: List[str] = ("H", "D", "Z", "F"),
        type: str = "variation",
        interval: str = "minute",
        urlTemplate: str = "https://geomag.usgs.gov/ws/data",
        urlInterval: int = -1,
        **kwargs,
    ):
        super().__init__(
            observatory=observatory,
            channels=channels,
            type=type,
            interval=interval,
            urlTemplate=urlTemplate,
            urlInterval=urlInterval,
        )
        self.data_format = "json"
        self.observatoryMetadata = ObservatoryMetadata()
        self.geomag_urls = [
            url.strip() for url in os.getenv("GEOMAG_WS_HOSTS", "usgs.gov").split(",")
        ]  # a csv list of urls like [subdomain.domain.tld,...,]

    def _get_url(
        self,
        observatory: str,
        date: dict[Literal["start", "end"], UTCDateTime],
        type: str = "variation",
        interval: str = "minute",
        channels: Optional[List[str]] = None,
    ) -> str:
        """
        Build the URL for the web service request by appending required query parameters.
        The WS API requires at minimum:
          - id (observatory code)
          - starttime (ISO format)
          - endtime (ISO format, computed as starttime + urlInterval seconds)
          - sampling_period (in seconds, derived from the interval)
          - type (data type)
          - elements (if channels are provided, as a comma separated string)
        """
        # Unpack start and end times.
        starttime = date["start"]
        endtime = date["end"]
        sampling_period_map = {
            "tenhertz": 0.1,
            "second": 1,
            "minute": 60,
            "hour": 3600,
            "day": 86400,
        }
        sampling_period = sampling_period_map.get(interval)
        if sampling_period is None:
            raise TimeseriesFactoryException("Unsupported interval: " + interval)

        if any(url in self.urlTemplate for url in self.geomag_urls):
            # TODO: Implement other webservices here, starting with public urls first.
            params = self.build_params_usgs(
                observatory,
                starttime,
                endtime,
                sampling_period,
                type,
                self.data_format,
                channels,
            )
        else:
            raise TimeseriesFactoryException(
                "Query Param builder not available for ws at: " + self.urlTemplate
            )
        query_string = urlencode(params)
        parsed_url = urlparse(self.urlTemplate)
        built_url = urlunparse(parsed_url._replace(query=query_string))
        return built_url

    def build_params_usgs(
        self,
        observatory: str,
        starttime: UTCDateTime,
        endtime: UTCDateTime,
        sampling_period: str,
        type: DataType,
        format: str,
        channels=None,
    ) -> Union[DataApiQuery, FilterApiQuery]:
        params: Union[DataApiQuery, FilterApiQuery] = {
            "id": observatory,
            "starttime": starttime.isoformat(),
            "type": type,
            "format": format,
        }
        parsed_url = urlparse(self.urlTemplate)
        if "/ws/data" in parsed_url.path:
            params["sampling_period"] = sampling_period
        else:
            params["output_sampling_period"] = sampling_period

        if isinstance(endtime, UTCDateTime):
            params["endtime"] = endtime.isoformat()

        if channels:
            params["elements"] = ",".join(channels)

        existing_params = parse_qs(parsed_url.query)
        if len(existing_params) > 0:
            existing_params = {
                k: ",".join(v) for k, v in existing_params.items()
            }  # parse_qs returns values as list, so make into csv
            params.update(existing_params)  # user provided query values take precedent.

        return params

    def parse_string(self, data: str, **kwargs) -> Stream:
        """
        Parse the JSON data returned from the geomag web service into an ObsPy Stream.
        """
        try:
            stream = None
            if any(
                url in self.urlTemplate for url in self.geomag_urls
            ):  # replace with gitlab variable
                stream = self.parse_geomag_ws(data)
            else:
                raise TimeseriesFactoryException(
                    "Parser not implemented for service at: " + self.urlTemplate
                )
            if stream is not None:
                self._set_metadata(stream)
            return stream
        except Exception as e:
            raise TimeseriesFactoryException("Error Loading JSON: " + str(e))

    def parse_geomag_ws(self, data) -> Stream:
        if self.data_format == "json":
            return IMFJSONParser.parse(data)
        else:
            raise TimeseriesFactoryException("data format not implemented")

    def _set_metadata(self, stream):
        for trace in stream:
            self.observatoryMetadata.set_metadata(
                trace.stats,
                trace.stats["station"],
                trace.stats["channel"],
                self.type,
                self.interval,
            )
