import pytest

from datetime import datetime, timezone
from obspy import UTCDateTime

from geomagio.api.db.models.metadata import (
    Metadata,
    MetadataTable,
    MetadataCategory,
)
from geomagio.api.db.models.metadata_history import (
    MetadataHistoryTable,
    MetadataHistory,
)


@pytest.fixture()
def valid_token():
    return {
        "access_token": "test_access_token",
        "token_type": "Bearer",
        "expires_in": 7200,
        "refresh_token": "test_refresh_toekn",
        "scope": "openid email profile",
        "created_at": 1733244369,
        "id_token": "test_id_token",
        "expires_at": 1733251569,
        "userinfo": {
            "iss": "http://test_url",
            "sub": "1234",
            "aud": "test_aud",
            "exp": 1733244489,
            "iat": 1733244369,
            "nonce": "test_nonce",
            "auth_time": 1733244368,
            "sub_legacy": "test_sub_legacy",
            "name": "test_user",
            "nickname": "test_user_nickname",
            "preferred_username": "test_preferred_username",
            "email": "test_email",
            "email_verified": True,
            "profile": "http://test_url/user",
            "picture": "http://picture_url",
            "groups_direct": ["group1", "group2"],
        },
    }


@pytest.fixture()
def valid_userinfo():
    return {
        "sub": "1234",
        "sub_legacy": "test_sub_legacy",
        "name": "test_user",
        "nickname": "test_user_nickname",
        "preferred_username": "test_preferred_username",
        "email": "test_email",
        "email_verified": True,
        "profile": "http://test_url/user",
        "picture": "http://picture_url",
        "groups": ["group1", "group2"],
    }


@pytest.fixture()
def valid_gitlab_user():
    return {
        "id": 1234,
        "email": "test_email",
        "name": "test_user",
        "username": "test_user_nickname",
        "avatar_url": "http://picture_url",
    }


@pytest.fixture()
def valid_session(valid_token, valid_userinfo):
    return {"token": valid_token, "user": valid_userinfo}


@pytest.fixture()
def valid_reading_metadata():
    return {
        "created_by": "test_user",
        "starttime": "2024-11-07T12:12:12Z",
        "endtime": "2024-11-07T20:20:20Z",
        "network": "NT",
        "station": "BOU",
        "category": "reading",
        "data_valid": True,
        "metadata": {
            "azimuth": 199.1383,
            "metadata": {
                "time": "2020-01-03T17:12:47Z",
                "station": "BOU",
                "observer": "Test Observer",
                "reviewed": True,
                "reviewer": "Test Reviewer",
                "mark_name": "AZ",
                "pier_name": "MainPCDCP",
                "theodolite": "108449",
                "electronics": "0110",
                "mark_azimuth": 199.1383,
                "pier_correction": -22,
            },
            "absolutes": [
                {
                    "shift": 0,
                    "valid": True,
                    "element": "D",
                    "endtime": "2020-01-03T17:16:21.000000Z",
                    "absolute": 8.3851056,
                    "baseline": 8.58571,
                    "starttime": "2020-01-03T17:12:47.000000Z",
                },
                {
                    "shift": 0,
                    "valid": True,
                    "element": "H",
                    "endtime": "2020-01-03T17:24:40.000000Z",
                    "absolute": 20728.0650365,
                    "baseline": -71.7177135,
                    "starttime": "2020-01-03T17:20:48.000000Z",
                },
                {
                    "shift": 0,
                    "valid": True,
                    "element": "Z",
                    "endtime": "2020-01-03T17:24:40.000000Z",
                    "absolute": 47450.1529433,
                    "baseline": 578.2041933,
                    "starttime": "2020-01-03T17:20:48.000000Z",
                },
            ],
            "hemisphere": 1,
            "diagnostics": None,
            "scale_value": None,
            "measurements": [
                {
                    "e": None,
                    "f": None,
                    "h": None,
                    "z": None,
                    "angle": 191.56666666666666,
                    "measurement_type": "FirstMarkDown",
                },
                {
                    "e": None,
                    "f": None,
                    "h": None,
                    "z": None,
                    "angle": 11.566666666666666,
                    "measurement_type": "FirstMarkUp",
                },
                {
                    "e": None,
                    "f": None,
                    "h": None,
                    "z": None,
                    "angle": 191.56666666666666,
                    "measurement_type": "SecondMarkDown",
                },
                {
                    "e": None,
                    "f": None,
                    "h": None,
                    "z": None,
                    "angle": 11.566666666666666,
                    "measurement_type": "SecondMarkUp",
                },
                {
                    "e": -72.242,
                    "f": 51801.81,
                    "h": 20800.329,
                    "z": 46871.49,
                    "angle": 270.71666666666664,
                    "measurement_type": "WestDown",
                    "time": "2024-11-07T12:12:12Z",
                },
                {
                    "e": -72.636,
                    "f": 51801.92,
                    "h": 20800.259,
                    "z": 46871.641,
                    "angle": 90.66666666666667,
                    "measurement_type": "EastDown",
                    "time": "2024-11-07T12:13:14Z",
                },
                {
                    "e": -72.657,
                    "f": 51801.82,
                    "h": 20800.259,
                    "z": 46871.521,
                    "angle": 90.93333333333334,
                    "measurement_type": "WestUp",
                    "time": "2024-11-07T13:13:11Z",
                },
                {
                    "e": -72.758,
                    "f": 51801.92,
                    "h": 20800.086,
                    "z": 46871.707,
                    "angle": 270.96666666666664,
                    "measurement_type": "EastUp",
                    "time": "2024-11-07T13:15:12Z",
                },
                {
                    "e": -72.898,
                    "f": 51801.89,
                    "h": 20799.796,
                    "z": 46871.802,
                    "angle": 246.38333333333333,
                    "measurement_type": "SouthDown",
                    "time": "2024-11-07T15:12:12Z",
                },
                {
                    "e": -72.8,
                    "f": 51802.01,
                    "h": 20799.852,
                    "z": 46871.919,
                    "angle": 66.38333333333334,
                    "measurement_type": "NorthUp",
                    "time": "2024-11-07T18:12:12Z",
                },
                {
                    "e": -72.775,
                    "f": 51802.01,
                    "h": 20799.668,
                    "z": 46871.997,
                    "angle": 113.58333333333333,
                    "measurement_type": "SouthUp",
                    "time": "2024-11-07T19:12:19Z",
                },
                {
                    "e": -72.813,
                    "f": 51802.14,
                    "h": 20799.815,
                    "z": 46872.077,
                    "angle": 293.5833333333333,
                    "measurement_type": "NorthDown",
                    "time": "2024-11-07T20:20:20Z",
                },
                {"measurement_type": "Meridian", "angle": 21},
            ],
            "pier_correction": -22,
        },
    }


@pytest.fixture()
def instrument_metadata():
    return Metadata(
        created_by="test_metadata.py",
        category=MetadataCategory.INSTRUMENT,
        network="NT",
        station="BDT",
        location="R0",
        channel="F",
        metadata={
            "type": "FGE",
            "channels": {
                "U": [{"channel": "U_Volt", "offset": 0, "scale": 313.2}],
                "V": [{"channel": "V_Volt", "offset": 0, "scale": 312.3}],
                "W": [{"channel": "W_Volt", "offset": 0, "scale": 312.0}],
            },
            "electronics": {
                "serial": "E0542",
                "x-scale": 313.2,
                "y-scale": 312.3,
                "z-scale": 312.0,
                "temperature-scale": 0.01,
            },
            "sensor": {
                "serial": "S0419",
                "x-constant": 36958,
                "y-constant": 36849,
                "z-constant": 36811,
            },
        },
        status="new",
    )


@pytest.fixture()
def metadata_with_times():
    return Metadata(
        created_by="test_metadata.py",
        created_time=UTCDateTime(2021, 1, 3, 17, 24, 40),
        updated_time=UTCDateTime(2021, 2, 3, 17, 24, 40),
        starttime=UTCDateTime(2020, 12, 3, 22, 43, 27),
        endtime=UTCDateTime(2020, 12, 3, 23, 43, 27),
        category=MetadataCategory.INSTRUMENT,
        network="NT",
        station="BOU",
        channel="U",
        metadata={
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        },
        status="updated",
    )


@pytest.fixture()
def metadata_with_datetimes():
    return Metadata(
        created_by="test_metadata.py",
        created_time=datetime(2021, 1, 3, 17, 24, 40, tzinfo=timezone.utc),
        updated_time=datetime(2021, 2, 3, 17, 24, 40, tzinfo=timezone.utc),
        starttime=datetime(2020, 12, 4, 22, 43, 27, tzinfo=timezone.utc),
        endtime=datetime(2020, 12, 4, 23, 43, 27, tzinfo=timezone.utc),
        category=MetadataCategory.INSTRUMENT,
        network="NT",
        station="BOU",
        location="R0",
        metadata={
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        },
        status="updated",
    )


@pytest.fixture()
def metadata_with_datetime_naive():
    return Metadata(
        created_time=datetime(2021, 1, 3, 17, 24, 40),
        updated_time=datetime(2021, 2, 3, 17, 24, 40),
        starttime=datetime(2020, 12, 5, 22, 43, 27),
        endtime=datetime(2020, 12, 5, 23, 43, 27),
        category=MetadataCategory.INSTRUMENT,
        network="NT",
        station="BOU",
        metadata={
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        },
        status="updated",
    )


@pytest.fixture()
def metadata_with_time_strings():
    return Metadata(
        created_by="test_metadata.py",
        created_time="2021-01-03T17:24:40",
        starttime="2020-12-03T22:43:27",
        endtime="2020-12-03T23:43:27",
        category=MetadataCategory.INSTRUMENT,
        network="NT",
        station="BOU",
        metadata={
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        },
    )


@pytest.fixture()
def observatory_metadata():
    return Metadata(
        created_by="test_metadata.py",
        starttime="2022-02-03T22:43:27",
        endtime="2022-02-03T23:43:27",
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "12"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def get_all_metadata(
    instrument_metadata,
    metadata_with_times,
    metadata_with_datetimes,
    metadata_with_datetime_naive,
    metadata_with_time_strings,
    observatory_metadata,
):
    return [
        instrument_metadata,
        metadata_with_times,
        metadata_with_datetimes,
        metadata_with_datetime_naive,
        metadata_with_time_strings,
        observatory_metadata,
    ]


@pytest.fixture()
def metadata_history_1():
    return MetadataHistoryTable(
        metadata_id=38742,
        created_by="test_metadata.py",
        starttime="2022-02-03T22:43:27",
        endtime="2022-02-03T23:43:27",
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "12"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def metadata_history_2():
    return MetadataHistoryTable(
        metadata_id=38742,
        created_by="test_metadata.py",
        updated_by="tester",
        updated_time="2023-01-21T03:30:43",
        starttime="2022-02-03T22:43:27",
        endtime="2022-02-03T23:43:27",
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def metadata_history_3():
    return MetadataHistoryTable(
        metadata_id=38742,
        created_by="test_metadata.py",
        updated_by="another_tester",
        updated_time="2023-01-22T13:30:43",
        starttime="2022-02-03T22:43:27",
        endtime="2022-02-03T23:43:27",
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 87.32,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def current_metadata_for_related_metadata_history():
    return MetadataTable(
        id=38742,
        created_by="test_metadata.py",
        created_time=datetime(2025, 1, 22, 13, 45, 27, tzinfo=timezone.utc),
        updated_by="another_tester",
        updated_time=datetime(2023, 6, 1, 13, 45, 27, tzinfo=timezone.utc),
        starttime=datetime(2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc),
        endtime=datetime(2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc),
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata_={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 87.32,
            "elevation": 10,
            "longitude": 202.234,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def get_all_related_metadata_history(
    metadata_history_1, metadata_history_2, metadata_history_3
):
    return [metadata_history_1, metadata_history_2, metadata_history_3]


@pytest.fixture()
def formatted_metadata_history_1():
    return MetadataHistory(
        metadata_id=38746,
        created_by="test_metadata.py",
        starttime=datetime(2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc),
        endtime=datetime(2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc),
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "12"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def formatted_metadata_history_2():
    return MetadataHistory(
        metadata_id=38746,
        created_by="test_metadata.py",
        updated_by="tester",
        updated_time=datetime(2023, 1, 21, 3, 30, 43, tzinfo=timezone.utc),
        starttime=datetime(2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc),
        endtime=datetime(2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc),
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def formatted_metadata_history_3():
    return MetadataHistory(
        metadata_id=38746,
        created_by="test_metadata.py",
        updated_by="another_tester",
        updated_time=datetime(2023, 1, 22, 3, 30, 43, tzinfo=timezone.utc),
        starttime=datetime(2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc),
        endtime=datetime(2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc),
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 87.32,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def formatted_metadata_for_formatted_metadata_history():
    return Metadata(
        id=38746,
        created_by="test_metadata.py",
        created_time=datetime(2025, 1, 22, 13, 45, 27, tzinfo=timezone.utc),
        updated_by="another_tester",
        updated_time=datetime(2023, 6, 1, 13, 45, 27, tzinfo=timezone.utc),
        starttime=datetime(2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc),
        endtime=datetime(2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc),
        category=MetadataCategory.OBSERVATORY,
        network="NT",
        station="BRW",
        metadata={
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 87.32,
            "elevation": 10,
            "longitude": 202.234,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        },
    )


@pytest.fixture()
def get_all_formatted_metadata_history(
    formatted_metadata_for_formatted_metadata_history,
    formatted_metadata_history_3,
    formatted_metadata_history_2,
    formatted_metadata_history_1,
):
    return [
        formatted_metadata_for_formatted_metadata_history,
        formatted_metadata_history_3,
        formatted_metadata_history_2,
        formatted_metadata_history_1,
    ]
