import pytest
import pydantic_core

from obspy import UTCDateTime
from datetime import datetime, timedelta, timezone
from sqlmodel import Session, delete, select

from geomagio.api.db.models.metadata import (
    Metadata,
    MetadataTable,
    MetadataCategory,
    MetadataQuery,
    MetadataUpdate,
)
from geomagio.api.db.models.metadata_history import MetadataHistoryTable
from geomagio.api.db.factories.MetadataDatabaseFactory import MetadataDatabaseFactory


def compare_metadata_objects(a: MetadataTable, b: Metadata) -> bool:
    if a.id != b.id:
        raise ValueError(
            "ids do not match. a.id = " + str(a.id) + " and b.id = " + str(b.id)
        )
    if a.created_by != b.created_by:
        raise ValueError(
            "created_by does not match. a.created_by = "
            + a.created_by
            + " and b.created_by = "
            + b.created_by
        )
    if a.created_time != b.created_time:
        raise ValueError(
            "created_time does not match. a.created_time = "
            + a.created_time
            + " and b.created_time = "
            + b.created_time
        )
    if a.updated_by != b.updated_by:
        raise ValueError(
            "updated_by does not match. a.updated_by = "
            + a.updated_by
            + " and b.updated_by = "
            + b.updated_by
        )
    if a.updated_time != b.updated_time:
        raise ValueError(
            "updated_time does not match. a.updated_time = "
            + a.updated_time
            + " and b.updated_time = "
            + b.updated_time
        )
    if a.starttime != b.starttime:
        raise ValueError(
            "starttime does not match. a.starttime = "
            + a.starttime
            + " and b.starttime = "
            + b.starttime
        )
    if a.endtime != b.endtime:
        raise ValueError(
            "endtime does not match. a.endtime = "
            + a.endtime
            + " and b.endtime = "
            + b.endtime
        )
    if a.network != b.network:
        raise ValueError(
            "network does not match. a.network = "
            + a.network
            + " and b.network = "
            + b.network
        )
    if a.station != b.station:
        raise ValueError(
            "station does not match. a.station = "
            + a.station
            + " and b.station = "
            + b.station
        )
    if a.channel != b.channel:
        raise ValueError(
            "channel does not match. a.channel = "
            + a.channel
            + " and b.channel = "
            + b.channel
        )
    if a.location != b.location:
        raise ValueError(
            "location does not match. a.location = "
            + a.location
            + " and b.location = "
            + b.location
        )
    if a.category != b.category:
        raise ValueError(
            "category does not match. a.category = "
            + a.category
            + " and b.category = "
            + b.category
        )
    if a.priority != b.priority:
        raise ValueError(
            "priority does not match. a.priority = "
            + a.priority
            + " and b.priority = "
            + b.priority
        )
    if a.data_valid != b.data_valid:
        raise ValueError(
            "data_valid does not match. a.data_valid = "
            + a.data_valid
            + " and b.data_valid = "
            + b.data_valid
        )
    if a.status != b.status:
        raise ValueError(
            "status does not match. a.status = "
            + a.status
            + " and b.status = "
            + b.status
        )
    if a.metadata_ != b.metadata:
        raise ValueError(
            "metadata does not match. a.metadata_ = "
            + a.metadata
            + " and b.metadata = "
            + b.metadata
        )
    if a.comment != b.comment:
        raise ValueError(
            "comment does not match. a.comment = "
            + a.comment
            + " and b.comment = "
            + b.comment
        )
    if a.review_comment != b.review_comment:
        raise ValueError(
            "review_comment does not match. a.review_comment = "
            + a.review_comment
            + " and b.review_comment = "
            + b.review_comment
        )
    return True


@pytest.mark.asyncio
async def test_create_metadata_defaults(test_engine, instrument_metadata):
    instrument_metadata.id = 123456
    returned_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(new_metadata=instrument_metadata)

    # get the newly saved record to verify it was created correctly
    with Session(test_engine) as test_db_session:
        saved_metadata = test_db_session.get(MetadataTable, returned_metadata.id)

        # assert the saved metadata fields are set correctly
        assert saved_metadata.created_by == "test_metadata.py"
        assert saved_metadata.category == MetadataCategory.INSTRUMENT
        assert saved_metadata.network == "NT"
        assert saved_metadata.station == "BDT"
        assert saved_metadata.metadata_ == {
            "type": "FGE",
            "channels": {
                "U": [{"channel": "U_Volt", "offset": 0, "scale": 313.2}],
                "V": [{"channel": "V_Volt", "offset": 0, "scale": 312.3}],
                "W": [{"channel": "W_Volt", "offset": 0, "scale": 312.0}],
            },
            "electronics": {
                "serial": "E0542",
                "x-scale": 313.2,
                "y-scale": 312.3,
                "z-scale": 312.0,
                "temperature-scale": 0.01,
            },
            "sensor": {
                "serial": "S0419",
                "x-constant": 36958,
                "y-constant": 36849,
                "z-constant": 36811,
            },
        }

        # assert created_time is within one hour of current time
        discard = timedelta(
            minutes=saved_metadata.created_time.minute,
            seconds=saved_metadata.created_time.second,
            microseconds=saved_metadata.created_time.microsecond,
        )
        saved_time = saved_metadata.created_time - discard

        now = UTCDateTime.now()
        current_time = datetime(
            year=now.year,
            month=now.month,
            day=now.day,
            hour=now.hour,
            tzinfo=timezone.utc,
        )

        assert current_time == saved_time

        # assert defaults are set correctly
        assert saved_metadata.priority == 1
        assert saved_metadata.data_valid == True
        assert saved_metadata.status == "new"

        # assert id is the database assigned id. not the input id
        assert saved_metadata.id != instrument_metadata.id

        # assert the metadata returned is the same as the saved metadata
        # print("returned_metadata:", returned_metadata)
        assert compare_metadata_objects(a=saved_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_create_metadata_with_starttime_and_endtime(
    test_engine, metadata_with_times
):
    returned_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(new_metadata=metadata_with_times)

    # get the newly saved record to verify it was created correctly
    with Session(test_engine) as test_db_session:
        saved_metadata = test_db_session.get(MetadataTable, returned_metadata.id)

        # assert the saved metadata fields are set correctly
        assert saved_metadata.created_by == "test_metadata.py"
        assert saved_metadata.created_time == datetime(
            2021, 1, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.updated_time == datetime(
            2021, 2, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.starttime == datetime(
            2020, 12, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.endtime == datetime(
            2020, 12, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.category == MetadataCategory.INSTRUMENT
        assert saved_metadata.network == "NT"
        assert saved_metadata.station == "BOU"
        assert saved_metadata.metadata_ == {
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        }

        # assert defaults are set correctly
        assert saved_metadata.priority == 1
        assert saved_metadata.data_valid == True
        assert saved_metadata.status == "updated"

        # assert the metadata returned is the same as the saved metadata
        assert compare_metadata_objects(a=saved_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_create_metadata_with_times_as_datetime(
    test_engine, metadata_with_datetimes
):
    returned_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(new_metadata=metadata_with_datetimes)

    # get the newly saved record to verify it was created correctly
    with Session(test_engine) as test_db_session:
        saved_metadata = test_db_session.get(MetadataTable, returned_metadata.id)

        # assert the saved metadata fields are set correctly
        assert saved_metadata.created_by == "test_metadata.py"
        assert saved_metadata.created_time == datetime(
            2021, 1, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.updated_time == datetime(
            2021, 2, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.starttime == datetime(
            2020, 12, 4, 22, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.endtime == datetime(
            2020, 12, 4, 23, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.category == MetadataCategory.INSTRUMENT
        assert saved_metadata.network == "NT"
        assert saved_metadata.station == "BOU"
        assert saved_metadata.metadata_ == {
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        }

        # assert defaults are set correctly
        assert saved_metadata.priority == 1
        assert saved_metadata.data_valid == True
        assert saved_metadata.status == "updated"

        # assert the metadata returned is the same as the saved metadata
        assert compare_metadata_objects(a=saved_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_create_metadata_with_times_as_naive_datetime(
    test_engine, metadata_with_datetime_naive
):
    returned_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(new_metadata=metadata_with_datetime_naive)

    # get the newly saved record to verify it was created correctly
    with Session(test_engine) as test_db_session:
        saved_metadata = test_db_session.get(MetadataTable, returned_metadata.id)

        # assert native datetimes are converted to aware datetimes by model_validate
        assert saved_metadata.created_time == datetime(
            2021, 1, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.updated_time == datetime(
            2021, 2, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.starttime == datetime(
            2020, 12, 5, 22, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.endtime == datetime(
            2020, 12, 5, 23, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.category == MetadataCategory.INSTRUMENT
        assert saved_metadata.network == "NT"
        assert saved_metadata.station == "BOU"
        assert saved_metadata.metadata_ == {
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        }

        # assert defaults are set correctly
        assert saved_metadata.priority == 1
        assert saved_metadata.data_valid == True
        assert saved_metadata.status == "updated"

        # assert the metadata returned is the same as the saved metadata
        assert compare_metadata_objects(a=saved_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_create_metadata_with_times_as_strings(
    test_engine, metadata_with_time_strings
):
    returned_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(new_metadata=metadata_with_time_strings)

    # get the newly saved record to verify it was created correctly
    with Session(test_engine) as test_db_session:
        saved_metadata = test_db_session.get(MetadataTable, returned_metadata.id)

        # assert the saved metadata fields are set correctly
        assert saved_metadata.created_by == "test_metadata.py"

        # assert native datetimes are converted to aware datetimes by model_validate
        assert saved_metadata.created_time == datetime(
            2021, 1, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.starttime == datetime(
            2020, 12, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.endtime == datetime(
            2020, 12, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.updated_time == None
        assert saved_metadata.category == MetadataCategory.INSTRUMENT
        assert saved_metadata.network == "NT"
        assert saved_metadata.station == "BOU"
        assert saved_metadata.metadata_ == {
            "type": "Narod",
            "channels": {
                "U": [
                    {"channel": "U_Volt", "offset": 0, "scale": 100},
                    {"channel": "U_Bin", "offset": 0, "scale": 500},
                ],
                "V": [
                    {"channel": "V_Volt", "offset": 0, "scale": 100},
                    {"channel": "V_Bin", "offset": 0, "scale": 500},
                ],
                "W": [
                    {"channel": "W_Volt", "offset": 0, "scale": 100},
                    {"channel": "W_Bin", "offset": 0, "scale": 500},
                ],
            },
        }

        # assert defaults are set correctly
        assert saved_metadata.priority == 1
        assert saved_metadata.data_valid == True
        assert saved_metadata.status == "new"

        # assert the metadata returned is the same as the saved metadata
        assert compare_metadata_objects(a=saved_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_create_metadata_no_nested_metadata(
    test_engine, metadata_with_time_strings
):

    metadata_with_time_strings.metadata = None
    returned_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(new_metadata=metadata_with_time_strings)

    # get the newly saved record to verify it was created correctly
    with Session(test_engine) as test_db_session:
        saved_metadata = test_db_session.get(MetadataTable, returned_metadata.id)

        # assert the saved metadata fields are set correctly
        assert saved_metadata.created_by == "test_metadata.py"

        # assert native datetimes are converted to aware datetimes by model_validate
        assert saved_metadata.created_time == datetime(
            2021, 1, 3, 17, 24, 40, tzinfo=timezone.utc
        )
        assert saved_metadata.starttime == datetime(
            2020, 12, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.endtime == datetime(
            2020, 12, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert saved_metadata.updated_time == None
        assert saved_metadata.category == MetadataCategory.INSTRUMENT
        assert saved_metadata.network == "NT"
        assert saved_metadata.station == "BOU"
        assert saved_metadata.metadata_ == None

        # assert defaults are set correctly
        assert saved_metadata.priority == 1
        assert saved_metadata.data_valid == True
        assert saved_metadata.status == "new"


@pytest.mark.asyncio
async def test_create_metadata_invalid_metadata(test_engine, instrument_metadata):
    instrument_metadata.starttime = "invalid starttime"

    try:
        await MetadataDatabaseFactory(engine=test_engine).create_metadata(
            new_metadata=instrument_metadata
        )
    except Exception as e:
        assert e is not None


@pytest.mark.asyncio
async def test_get_metadata_with_starttime(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        category=MetadataCategory.INSTRUMENT, starttime="2019-01-01T00:00:00"
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 5 metadata rows to be returned. 4 with the endtime after the query starttime and 1 with no endtime
    assert len(results) == 5

    for result in results:
        assert result.category == MetadataCategory.INSTRUMENT
        # assert metadata is set correctly and not as metadata_
        assert result.metadata["type"] != None


@pytest.mark.asyncio
async def test_get_metadata_with_endtime(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        category=MetadataCategory.INSTRUMENT, endtime="2020-12-04T00:00:00"
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 3 metadata rows to be returned. 2 with the starttime before the query endtime and 1 with no endtime
    assert len(results) == 3

    for result in results:
        assert result.category == MetadataCategory.INSTRUMENT
        # assert metadata is set correctly and not as metadata_
        assert result.metadata["type"] != None


@pytest.mark.asyncio
async def test_get_metadata_with_channel(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        category=MetadataCategory.INSTRUMENT, endtime="2020-12-04T00:00:00", channel="F"
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    assert len(results) == 1
    assert results[0].channel == "F"
    assert results[0].metadata["type"] == "FGE"


@pytest.mark.asyncio
async def test_get_metadata_with_location(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        category=MetadataCategory.INSTRUMENT,
        endtime="2020-12-04T00:00:00",
        location="R0",
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    assert len(results) == 1
    assert results[0].location == "R0"
    assert results[0].metadata["type"] == "FGE"


@pytest.mark.asyncio
async def test_get_metadata_with_created_by(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        created_by="test_metadata.py",
        category=MetadataCategory.INSTRUMENT,
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    assert len(results) == 4
    for result in results:
        assert result.category == MetadataCategory.INSTRUMENT
        # assert all results have the correct value for created_by
        assert result.created_by == "test_metadata.py"


@pytest.mark.asyncio
async def test_get_metadata_with_single_status(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        category=MetadataCategory.INSTRUMENT,
        starttime="2019-01-01T00:00:00",
        status=["new"],
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 2 metadata rows to be returned. the 1st object and the 5th where status is set to new by default
    assert len(results) == 2

    for result in results:
        assert result.category == MetadataCategory.INSTRUMENT
        assert result.status == "new"
        assert result.metadata["type"] != None


@pytest.mark.asyncio
async def test_get_metadata_with_multiple_statuses(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        category=MetadataCategory.INSTRUMENT,
        starttime="2019-01-01T00:00:00",
        status=["new", "updated"],
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 5 metadata rows to be returned. all of the instrument metadata has a "new" or "updated" status and passes the starttime conditional
    assert len(results) == 5

    for result in results:
        assert result.category == MetadataCategory.INSTRUMENT
        assert result.status == "new" or result.status == "updated"
        assert result.metadata["type"] != None


@pytest.mark.asyncio
async def test_get_metadata_with_created_before(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        created_before=UTCDateTime(2025, 1, 1),
        category=MetadataCategory.INSTRUMENT,
        starttime="2019-01-01T00:00:00",
        status=["new", "updated"],
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 4 metadata rows to be returned. all of the instrument metadata except for the first one
    assert len(results) == 4

    for result in results:
        assert result.category == MetadataCategory.INSTRUMENT
        assert result.status == "new" or result.status == "updated"
        assert result.metadata["type"] == "Narod"


@pytest.mark.asyncio
async def test_get_metadata_with_created_after(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        created_after=UTCDateTime(2019, 1, 1),
        category=MetadataCategory.INSTRUMENT,
        starttime="2019-01-01T00:00:00",
        status=["new", "updated"],
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 5 all of the instrument metadata rows to be returned
    assert len(results) == 5

    for result in results:
        assert result.category == MetadataCategory.INSTRUMENT
        assert result.status == "new" or result.status == "updated"
        assert result.metadata["type"] != None


@pytest.mark.asyncio
async def test_get_metadata_returns_one_in_list(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(
        created_after=UTCDateTime(2019, 1, 1), category=MetadataCategory.OBSERVATORY
    )

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 1 metadata observatory object to be returned as a list
    assert len(results) == 1
    assert results[0].category == MetadataCategory.OBSERVATORY
    assert results[0].metadata == {
        "id": "BRW",
        "name": "Barrow",
        "marks": [{"name": "m", "azimuth": "1"}],
        "piers": [{"name": "p", "correction": "12"}],
        "agency": "USGS",
        "latitude": 71.322,
        "elevation": 10,
        "longitude": 203.378,
        "agency_name": "United States Geological Survey (USGS)",
        "electronics": ["1"],
        "theodolites": ["12"],
        "declination_base": 10589,
        "sensor_orientation": "HDZF",
    }


@pytest.mark.asyncio
async def test_get_metadata_returns_empty_list(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    query = MetadataQuery(category=MetadataCategory.READING)

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # assert that it returns an empty list and does not error
    assert len(results) == 0


@pytest.mark.asyncio
async def test_get_metadata_no_nested_metadata(test_engine, instrument_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a metadata object with no nested metadata
    instrument_metadata.metadata = None
    empty_nested_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(instrument_metadata)

    query = MetadataQuery(id=empty_nested_metadata.id)

    results = await MetadataDatabaseFactory(engine=test_engine).get_metadata(
        query=query
    )

    # expect 1 metadata instrument object to be returned as a list with no nested metadata
    assert len(results) == 1
    assert results[0].category == MetadataCategory.INSTRUMENT
    assert results[0].metadata == None


@pytest.mark.asyncio
async def test_get_metadata_by_id(test_engine, get_all_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataDatabaseFactory(engine=test_engine).batch_create_metadata(
        get_all_metadata
    )

    result = await MetadataDatabaseFactory(engine=test_engine).get_metadata_by_id(3)
    assert result.id == 3
    assert result.created_time == datetime(2021, 1, 3, 17, 24, 40, tzinfo=timezone.utc)
    assert result.updated_time == datetime(2021, 2, 3, 17, 24, 40, tzinfo=timezone.utc)
    assert result.starttime == datetime(2020, 12, 4, 22, 43, 27, tzinfo=timezone.utc)
    assert result.endtime == datetime(2020, 12, 4, 23, 43, 27, tzinfo=timezone.utc)
    assert result.category == MetadataCategory.INSTRUMENT
    assert result.network == "NT"
    assert result.station == "BOU"
    assert result.location == "R0"
    assert result.metadata == {
        "type": "Narod",
        "channels": {
            "U": [
                {"channel": "U_Volt", "offset": 0, "scale": 100},
                {"channel": "U_Bin", "offset": 0, "scale": 500},
            ],
            "V": [
                {"channel": "V_Volt", "offset": 0, "scale": 100},
                {"channel": "V_Bin", "offset": 0, "scale": 500},
            ],
            "W": [
                {"channel": "W_Volt", "offset": 0, "scale": 100},
                {"channel": "W_Bin", "offset": 0, "scale": 500},
            ],
        },
    }


@pytest.mark.asyncio
async def test_get_metadata_by_id_no_nested_metadata(test_engine, instrument_metadata):
    # clear in memory Metadata table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a metadata object that has no nested metadata
    instrument_metadata.metadata = None
    empty_nested_metadata = await MetadataDatabaseFactory(
        engine=test_engine
    ).create_metadata(instrument_metadata)

    result = await MetadataDatabaseFactory(engine=test_engine).get_metadata_by_id(
        empty_nested_metadata.id
    )
    assert result.id == empty_nested_metadata.id
    assert result.network == "NT"
    assert result.station == "BDT"
    assert result.location == "R0"
    assert result.metadata == None


@pytest.mark.asyncio
async def test_get_metadata_by_id_no_metadata_returned(test_engine):
    result = await MetadataDatabaseFactory(engine=test_engine).get_metadata_by_id(12345)

    # assert the result is None and not an error
    assert result == None


@pytest.mark.asyncio
async def test_get_metadata_by_id_invalid(test_engine):
    try:
        await MetadataDatabaseFactory(engine=test_engine).get_metadata_by_id("3")
    except Exception as e:
        assert str(e) == "expected an int"


def compare_metadata_and_metadatahistory_objects(
    a: Metadata, b: MetadataHistoryTable
) -> bool:
    if a.id != b.metadata_id:
        raise ValueError(
            "ids do not match. a.id = "
            + str(a.id)
            + " and b.metadata_id = "
            + str(b.metadata_id)
        )
    if a.created_by != b.created_by:
        raise ValueError(
            "created_by does not match. a.created_by = "
            + a.created_by
            + " and b.created_by = "
            + b.created_by
        )
    if a.created_time != b.created_time:
        raise ValueError(
            "created_time does not match. a.created_time = "
            + a.created_time
            + " and b.created_time = "
            + b.created_time
        )
    if a.updated_by != b.updated_by:
        raise ValueError(
            "updated_by does not match. a.updated_by = "
            + a.updated_by
            + " and b.updated_by = "
            + b.updated_by
        )
    if a.updated_time != b.updated_time:
        raise ValueError(
            "updated_time does not match. a.updated_time = "
            + a.updated_time
            + " and b.updated_time = "
            + b.updated_time
        )
    if a.starttime != b.starttime:
        raise ValueError(
            "starttime does not match. a.starttime = "
            + a.starttime
            + " and b.starttime = "
            + b.starttime
        )
    if a.endtime != b.endtime:
        raise ValueError(
            "endtime does not match. a.endtime = "
            + a.endtime
            + " and b.endtime = "
            + b.endtime
        )
    if a.network != b.network:
        raise ValueError(
            "network does not match. a.network = "
            + a.network
            + " and b.network = "
            + b.network
        )
    if a.station != b.station:
        raise ValueError(
            "station does not match. a.station = "
            + a.station
            + " and b.station = "
            + b.station
        )
    if a.channel != b.channel:
        raise ValueError(
            "channel does not match. a.channel = "
            + a.channel
            + " and b.channel = "
            + b.channel
        )
    if a.location != b.location:
        raise ValueError(
            "location does not match. a.location = "
            + a.location
            + " and b.location = "
            + b.location
        )
    if a.category != b.category:
        raise ValueError(
            "category does not match. a.category = "
            + a.category
            + " and b.category = "
            + b.category
        )
    if a.priority != b.priority:
        raise ValueError(
            "priority does not match. a.priority = "
            + a.priority
            + " and b.priority = "
            + b.priority
        )
    if a.data_valid != b.data_valid:
        raise ValueError(
            "data_valid does not match. a.data_valid = "
            + a.data_valid
            + " and b.data_valid = "
            + b.data_valid
        )
    if a.status != b.status:
        raise ValueError(
            "status does not match. a.status = "
            + a.status
            + " and b.status = "
            + b.status
        )
    if a.metadata != b.metadata_:
        raise ValueError(
            "metadata does not match. a.metadata = "
            + a.metadata
            + " and b.metadata_ = "
            + b.metadata_
        )
    if a.comment != b.comment:
        raise ValueError(
            "comment does not match. a.comment = "
            + a.comment
            + " and b.comment = "
            + b.comment
        )
    if a.review_comment != b.review_comment:
        raise ValueError(
            "review_comment does not match. a.review_comment = "
            + a.review_comment
            + " and b.review_comment = "
            + b.review_comment
        )
    return True


@pytest.mark.asyncio
async def test_update_metadata(test_engine, observatory_metadata):
    with Session(test_engine) as test_db_session:
        # clear in memory MetadataHistoryTable table
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

        # create a metadata row to update
        created_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).create_metadata(observatory_metadata)

        updated_metadata = MetadataUpdate(
            id=created_metadata.id, comment="Adding a comment", status="reviewed"
        )

        returned_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).update_metadata(updated_metadata=updated_metadata, updated_by="tester")

        # get metadatahistory row
        statement = select(MetadataHistoryTable).where(
            MetadataHistoryTable.metadata_id == created_metadata.id
        )
        results = test_db_session.exec(statement)
        metadatahistory = results.one()

        # assert the original created metadata saved to the metadatahistory table is identical to the original created metadata aside from id and metadataid
        assert compare_metadata_and_metadatahistory_objects(
            a=created_metadata, b=metadatahistory
        )

        # get metadata row to assert it was updated with the new values
        newly_updated_metadata = test_db_session.get(MetadataTable, created_metadata.id)

        # assert values not included in the MetadataUpdate object are not changed
        assert newly_updated_metadata.id == created_metadata.id
        assert newly_updated_metadata.created_by == created_metadata.created_by
        assert newly_updated_metadata.created_time == created_metadata.created_time
        assert newly_updated_metadata.starttime == created_metadata.starttime
        assert newly_updated_metadata.endtime == created_metadata.endtime
        assert newly_updated_metadata.network == created_metadata.network
        assert newly_updated_metadata.station == created_metadata.station
        assert newly_updated_metadata.channel == created_metadata.channel
        assert newly_updated_metadata.location == created_metadata.location
        assert newly_updated_metadata.category == created_metadata.category
        assert newly_updated_metadata.priority == created_metadata.priority
        assert newly_updated_metadata.data_valid == created_metadata.data_valid
        assert newly_updated_metadata.metadata_ == created_metadata.metadata
        assert newly_updated_metadata.review_comment == created_metadata.review_comment

        # assert updated_by is set correctly
        assert newly_updated_metadata.updated_by == "tester"
        discard = timedelta(
            minutes=newly_updated_metadata.created_time.minute,
            seconds=newly_updated_metadata.created_time.second,
            microseconds=newly_updated_metadata.created_time.microsecond,
        )
        saved_time = newly_updated_metadata.created_time - discard

        now = UTCDateTime.now()
        current_time = datetime(
            year=now.year,
            month=now.month,
            day=now.day,
            hour=now.hour,
            tzinfo=timezone.utc,
        )

        # assert updated_time is set to within one minute of now
        assert current_time == saved_time

        # assert the values included in the MetadataUpdate object are changed to the new values
        assert newly_updated_metadata.status == "reviewed"
        assert newly_updated_metadata.comment == "Adding a comment"

        # assert the returned updated metadata is the same as the metadata table row
        assert compare_metadata_objects(a=newly_updated_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_update_metadata_original_object_empty_nested_metadata(
    test_engine, observatory_metadata
):
    with Session(test_engine) as test_db_session:
        # clear in memory MetadataHistoryTable table
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

        # create a metadata row to update with no nested metadata
        observatory_metadata.metadata = None
        created_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).create_metadata(observatory_metadata)

        updated_metadata = MetadataUpdate(
            id=created_metadata.id, comment="Adding a comment", status="reviewed"
        )

        returned_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).update_metadata(updated_metadata=updated_metadata, updated_by="tester")

        # get metadatahistory row
        statement = select(MetadataHistoryTable).where(
            MetadataHistoryTable.metadata_id == created_metadata.id
        )
        results = test_db_session.exec(statement)
        metadatahistory = results.one()

        # assert the original created metadata saved to the metadatahistory table is identical to the original created metadata aside from id and metadataid
        assert compare_metadata_and_metadatahistory_objects(
            a=created_metadata, b=metadatahistory
        )

        # get metadata row to assert it was updated with the new values
        newly_updated_metadata = test_db_session.get(MetadataTable, created_metadata.id)

        # assert values not included in the MetadataUpdate object are not changed
        assert newly_updated_metadata.id == created_metadata.id
        assert newly_updated_metadata.created_by == created_metadata.created_by
        assert newly_updated_metadata.created_time == created_metadata.created_time
        assert newly_updated_metadata.starttime == created_metadata.starttime
        assert newly_updated_metadata.endtime == created_metadata.endtime
        assert newly_updated_metadata.network == created_metadata.network
        assert newly_updated_metadata.station == created_metadata.station
        assert newly_updated_metadata.channel == created_metadata.channel
        assert newly_updated_metadata.location == created_metadata.location
        assert newly_updated_metadata.category == created_metadata.category
        assert newly_updated_metadata.priority == created_metadata.priority
        assert newly_updated_metadata.data_valid == created_metadata.data_valid
        assert newly_updated_metadata.metadata_ == None
        assert newly_updated_metadata.review_comment == created_metadata.review_comment

        # assert updated_by is set correctly
        assert newly_updated_metadata.updated_by == "tester"
        discard = timedelta(
            minutes=newly_updated_metadata.created_time.minute,
            seconds=newly_updated_metadata.created_time.second,
            microseconds=newly_updated_metadata.created_time.microsecond,
        )
        saved_time = newly_updated_metadata.created_time - discard

        now = UTCDateTime.now()
        current_time = datetime(
            year=now.year,
            month=now.month,
            day=now.day,
            hour=now.hour,
            tzinfo=timezone.utc,
        )

        # assert updated_time is set to within one minute of now
        assert current_time == saved_time

        # assert the values included in the MetadataUpdate object are changed to the new values
        assert newly_updated_metadata.status == "reviewed"
        assert newly_updated_metadata.comment == "Adding a comment"

        # assert the returned updated metadata is the same as the metadata table row
        assert compare_metadata_objects(a=newly_updated_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_update_metadata_whole_metadata_object(test_engine, observatory_metadata):
    with Session(test_engine) as test_db_session:
        # clear in memory MetadataHistoryTable table
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

        # create a metadata row to update
        created_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).create_metadata(observatory_metadata)

        # send the whole object and update the comment and status
        original_values = created_metadata.model_dump()

        updated_metadata = MetadataUpdate(**original_values)
        updated_metadata.comment = "Adding a comment"
        updated_metadata.status = "reviewed"

        returned_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).update_metadata(updated_metadata=updated_metadata, updated_by="tester")

        # get metadatahistory row
        statement = select(MetadataHistoryTable).where(
            MetadataHistoryTable.metadata_id == created_metadata.id
        )
        results = test_db_session.exec(statement)
        metadatahistory = results.one()

        # assert the original created metadata saved to the metadatahistory table is identical to the original created metadata aside from id and metadataid
        assert compare_metadata_and_metadatahistory_objects(
            a=created_metadata, b=metadatahistory
        )

        # get metadata row to assert it was updated with the new values
        newly_updated_metadata = test_db_session.get(MetadataTable, created_metadata.id)

        # assert values not included in the MetadataUpdate object are not changed
        assert newly_updated_metadata.id == created_metadata.id
        assert newly_updated_metadata.created_by == created_metadata.created_by
        assert newly_updated_metadata.created_time == created_metadata.created_time
        assert newly_updated_metadata.starttime == created_metadata.starttime
        assert newly_updated_metadata.endtime == created_metadata.endtime
        assert newly_updated_metadata.network == created_metadata.network
        assert newly_updated_metadata.station == created_metadata.station
        assert newly_updated_metadata.channel == created_metadata.channel
        assert newly_updated_metadata.location == created_metadata.location
        assert newly_updated_metadata.category == created_metadata.category
        assert newly_updated_metadata.priority == created_metadata.priority
        assert newly_updated_metadata.data_valid == created_metadata.data_valid
        assert newly_updated_metadata.metadata_ == created_metadata.metadata
        assert newly_updated_metadata.review_comment == created_metadata.review_comment

        # assert updated_by is set correctly
        assert newly_updated_metadata.updated_by == "tester"
        discard = timedelta(
            minutes=newly_updated_metadata.created_time.minute,
            seconds=newly_updated_metadata.created_time.second,
            microseconds=newly_updated_metadata.created_time.microsecond,
        )
        saved_time = newly_updated_metadata.created_time - discard

        now = UTCDateTime.now()
        current_time = datetime(
            year=now.year,
            month=now.month,
            day=now.day,
            hour=now.hour,
            tzinfo=timezone.utc,
        )

        # assert updated_time is set to within one minute of now
        assert current_time == saved_time

        # assert the values included in the MetadataUpdate object are changed to the new values
        assert newly_updated_metadata.status == "reviewed"
        assert newly_updated_metadata.comment == "Adding a comment"

        # assert the returned updated metadata is the same as the metadata table row
        assert compare_metadata_objects(a=newly_updated_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_update_metadata_remove_field(test_engine, observatory_metadata):
    with Session(test_engine) as test_db_session:
        # clear in memory MetadataHistoryTable table
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

        # create a metadata row to update
        observatory_metadata.comment = "to be deleted"
        created_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).create_metadata(observatory_metadata)

        updated_metadata = MetadataUpdate(
            id=created_metadata.id, comment=None, status="reviewed"
        )

        returned_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).update_metadata(updated_metadata=updated_metadata, updated_by="tester")

        # get metadatahistory row
        statement = select(MetadataHistoryTable).where(
            MetadataHistoryTable.metadata_id == created_metadata.id
        )
        results = test_db_session.exec(statement)
        metadatahistory = results.one()

        # assert the original created metadata saved to the metadatahistory table is identical to the original created metadata aside from id and metadataid
        assert compare_metadata_and_metadatahistory_objects(
            a=created_metadata, b=metadatahistory
        )

        # get metadata row to assert it was updated with the new values
        newly_updated_metadata = test_db_session.get(MetadataTable, created_metadata.id)

        # assert values not included in the MetadataUpdate object are not changed
        assert newly_updated_metadata.id == created_metadata.id
        assert newly_updated_metadata.created_by == created_metadata.created_by
        assert newly_updated_metadata.created_time == created_metadata.created_time
        assert newly_updated_metadata.starttime == created_metadata.starttime
        assert newly_updated_metadata.endtime == created_metadata.endtime
        assert newly_updated_metadata.network == created_metadata.network
        assert newly_updated_metadata.station == created_metadata.station
        assert newly_updated_metadata.channel == created_metadata.channel
        assert newly_updated_metadata.location == created_metadata.location
        assert newly_updated_metadata.category == created_metadata.category
        assert newly_updated_metadata.priority == created_metadata.priority
        assert newly_updated_metadata.data_valid == created_metadata.data_valid
        assert newly_updated_metadata.metadata_ == created_metadata.metadata
        assert newly_updated_metadata.review_comment == created_metadata.review_comment

        # assert updated_by is set correctly
        assert newly_updated_metadata.updated_by == "tester"
        discard = timedelta(
            minutes=newly_updated_metadata.created_time.minute,
            seconds=newly_updated_metadata.created_time.second,
            microseconds=newly_updated_metadata.created_time.microsecond,
        )
        saved_time = newly_updated_metadata.created_time - discard

        now = UTCDateTime.now()
        current_time = datetime(
            year=now.year,
            month=now.month,
            day=now.day,
            hour=now.hour,
            tzinfo=timezone.utc,
        )

        # assert updated_time is set to within one minute of now
        assert current_time == saved_time

        # assert the comment is removed
        assert newly_updated_metadata.status == "reviewed"
        assert newly_updated_metadata.comment == None

        # assert the returned updated metadata is the same as the metadata table row
        assert compare_metadata_objects(a=newly_updated_metadata, b=returned_metadata)


@pytest.mark.asyncio
async def test_update_metadata_multiple_updates(test_engine, observatory_metadata):
    with Session(test_engine) as test_db_session:
        # clear in memory MetadataHistoryTable table
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

        # create a metadata row to update
        created_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).create_metadata(observatory_metadata)

        updated_metadata = MetadataUpdate(
            id=created_metadata.id,
            metadata={
                "id": "BRW",
                "name": "Barrow",
                "marks": [{"name": "m", "azimuth": "1"}],
                "piers": [{"name": "p", "correction": "12"}],
                "agency": "USGS",
                "latitude": 100.00,
                "elevation": 10,
                "longitude": 203.378,
                "agency_name": "United States Geological Survey (USGS)",
                "electronics": ["1"],
                "theodolites": ["12"],
                "declination_base": 10589,
                "sensor_orientation": "HDZF",
            },
            status="updated",
        )

        returned_metadata = await MetadataDatabaseFactory(
            engine=test_engine
        ).update_metadata(updated_metadata=updated_metadata, updated_by="tester")

        # get metadatahistory row
        statement = select(MetadataHistoryTable).where(
            MetadataHistoryTable.metadata_id == created_metadata.id
        )
        results = test_db_session.exec(statement)
        metadatahistory = results.one()

        # assert the original created metadata saved to the metadatahistory table is identical to the original created metadata aside from id and metadataid
        assert compare_metadata_and_metadatahistory_objects(
            a=created_metadata, b=metadatahistory
        )

        # get metadata row to assert it was updated with the new values
        newly_updated_metadata = test_db_session.get(MetadataTable, created_metadata.id)

        # assert values not included in the MetadataUpdate object are not changed
        assert newly_updated_metadata.id == created_metadata.id
        assert newly_updated_metadata.created_by == created_metadata.created_by
        assert newly_updated_metadata.created_time == created_metadata.created_time
        assert newly_updated_metadata.starttime == created_metadata.starttime
        assert newly_updated_metadata.endtime == created_metadata.endtime
        assert newly_updated_metadata.network == created_metadata.network
        assert newly_updated_metadata.station == created_metadata.station
        assert newly_updated_metadata.channel == created_metadata.channel
        assert newly_updated_metadata.location == created_metadata.location
        assert newly_updated_metadata.category == created_metadata.category
        assert newly_updated_metadata.priority == created_metadata.priority
        assert newly_updated_metadata.data_valid == created_metadata.data_valid
        assert newly_updated_metadata.comment == created_metadata.comment
        assert newly_updated_metadata.review_comment == created_metadata.review_comment

        # assert updated_by is set correctly
        assert newly_updated_metadata.updated_by == "tester"
        discard = timedelta(
            minutes=newly_updated_metadata.created_time.minute,
            seconds=newly_updated_metadata.created_time.second,
            microseconds=newly_updated_metadata.created_time.microsecond,
        )
        saved_time = newly_updated_metadata.created_time - discard

        now = UTCDateTime.now()
        current_time = datetime(
            year=now.year,
            month=now.month,
            day=now.day,
            hour=now.hour,
            tzinfo=timezone.utc,
        )

        # assert updated_time is set to within one minute of now
        assert current_time == saved_time

        # assert the values included in the MetadataUpdate object are changed to the new values
        assert newly_updated_metadata.status == "updated"
        assert newly_updated_metadata.metadata_ == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "12"}],
            "agency": "USGS",
            "latitude": 100.00,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }

        # assert the returned updated metadata is the same as the metadata table row
        assert compare_metadata_objects(a=newly_updated_metadata, b=returned_metadata)

        # update the same metadata again
        second_update = MetadataUpdate(id=created_metadata.id, status="reviewed")

        second_returned = await MetadataDatabaseFactory(
            engine=test_engine
        ).update_metadata(updated_metadata=second_update, updated_by="different tester")

        # get the associated metadatahistory rows
        second_statement = select(MetadataHistoryTable).where(
            MetadataHistoryTable.metadata_id == created_metadata.id
        )
        second_results = test_db_session.exec(second_statement)
        metadatahistory_rows = second_results.all()

        assert len(metadatahistory_rows) == 2
        for row in metadatahistory_rows:
            assert row.metadata_id == created_metadata.id
            # assert the values that weren't updated didn't change
            assert row.created_by == created_metadata.created_by
            assert row.created_time == created_metadata.created_time
            assert row.starttime == created_metadata.starttime
            assert row.endtime == created_metadata.endtime
            assert row.network == created_metadata.network
            assert row.station == created_metadata.station
            assert row.channel == created_metadata.channel
            assert row.location == created_metadata.location
            assert row.category == created_metadata.category
            assert row.priority == created_metadata.priority
            assert row.data_valid == created_metadata.data_valid
            assert row.review_comment == created_metadata.review_comment

        assert metadatahistory_rows[0].status == "new"
        assert metadatahistory_rows[1].status == "updated"

        assert metadatahistory_rows[0].metadata_ == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "12"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }
        assert metadatahistory_rows[1].metadata_ == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "12"}],
            "agency": "USGS",
            "latitude": 100.00,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }

        assert metadatahistory_rows[1].updated_by == "tester"

        assert second_returned.updated_by == "different tester"
        assert second_returned.status == "reviewed"


@pytest.mark.asyncio
async def test_update_metadata_invalid_id(test_engine):
    updated_metadata = MetadataUpdate(
        id=476487562, comment="Adding a comment", status="reviewed"
    )

    try:
        await MetadataDatabaseFactory(engine=test_engine).update_metadata(
            updated_metadata=updated_metadata, updated_by="tester"
        )
    except Exception as e:
        assert str(e) == "metadata not found"
