import pytest

from datetime import datetime, timezone
from sqlmodel import Session, delete

from geomagio.api.db.models.metadata import (
    MetadataTable,
    MetadataQuery,
    MetadataCategory,
)
from geomagio.api.db.models.metadata_history import MetadataHistoryTable
from geomagio.api.db.factories.MetadataHistoryDatabaseFactory import (
    MetadataHistoryDatabaseFactory,
)


@pytest.mark.asyncio
async def test_get_metadata_history(test_engine, get_all_related_metadata_history):
    # clear in memory metadata_history table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata_history objects to the test database
    await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).batch_create_metadata_history(get_all_related_metadata_history)

    query = MetadataQuery(
        category=MetadataCategory.OBSERVATORY, starttime="2019-01-01T00:00:00"
    )

    results = await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).get_metadata_history(query=query)
    assert len(results) == 3

    for result in results:
        assert result.category == MetadataCategory.OBSERVATORY
        assert result.metadata["name"] == "Barrow"


@pytest.mark.asyncio
async def test_get_metadata_history_returns_one(test_engine, metadata_history_1):
    # clear in memory metadata_history table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a metadata_history object to the test database
    await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).batch_create_metadata_history([metadata_history_1])

    query = MetadataQuery(
        category=MetadataCategory.OBSERVATORY, starttime="2019-01-01T00:00:00"
    )

    results = await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).get_metadata_history(query=query)
    assert len(results) == 1
    assert results[0].category == MetadataCategory.OBSERVATORY
    assert results[0].metadata["name"] == "Barrow"


@pytest.mark.asyncio
async def test_get_metadata_history_returns_none(test_engine):
    # clear in memory metadata_history table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    query = MetadataQuery(
        category=MetadataCategory.OBSERVATORY, starttime="2019-01-01T00:00:00"
    )

    results = await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).get_metadata_history(query=query)

    # assert this returns an empty list and does not error
    assert len(results) == 0


@pytest.mark.asyncio
async def test_get_metadata_history_by_id(
    test_engine, get_all_related_metadata_history
):
    # clear in memory metadata_history table
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.commit()

    # save a bunch of metadata to the test database
    await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).batch_create_metadata_history(get_all_related_metadata_history)

    result = await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).get_metadata_history_by_id(2)
    assert result.id == 2
    assert result.metadata_id == 38742
    assert result.created_by == "test_metadata.py"
    assert result.updated_by == "tester"
    assert result.updated_time == datetime(2023, 1, 21, 3, 30, 43, tzinfo=timezone.utc)
    assert result.starttime == datetime(2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc)
    assert result.endtime == datetime(2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc)
    assert result.category == MetadataCategory.OBSERVATORY
    assert result.network == "NT"
    assert result.station == "BRW"
    assert result.metadata == {
        "id": "BRW",
        "name": "Barrow",
        "marks": [{"name": "m", "azimuth": "1"}],
        "piers": [{"name": "p", "correction": "22"}],
        "agency": "USGS",
        "latitude": 71.322,
        "elevation": 10,
        "longitude": 203.378,
        "agency_name": "United States Geological Survey (USGS)",
        "electronics": ["1"],
        "theodolites": ["12"],
        "declination_base": 10589,
        "sensor_orientation": "HDZF",
    }


@pytest.mark.asyncio
async def test_get_metadata_by_id_no_metadata_returned(test_engine):
    result = await MetadataHistoryDatabaseFactory(
        engine=test_engine
    ).get_metadata_history_by_id(12345)

    # assert the result is None and not an error
    assert result == None


@pytest.mark.asyncio
async def test_get_metadata_by_id_invalid(test_engine):
    try:
        await MetadataHistoryDatabaseFactory(
            engine=test_engine
        ).get_metadata_history_by_id("3")
    except Exception as e:
        assert str(e) == "expected an int"


@pytest.mark.asyncio
async def test_get_metadata_history_by_metadata_id(
    test_engine,
    current_metadata_for_related_metadata_history,
    get_all_related_metadata_history,
):
    # clear metadata and metadata_history tables
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        history_statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.exec(history_statement)
        test_db_session.commit()

        # save related metadata directly to db to make sure we're setting the right id
        test_db_session.add(current_metadata_for_related_metadata_history)
        test_db_session.commit()

        # save related metadata_history
        await MetadataHistoryDatabaseFactory(
            engine=test_engine
        ).batch_create_metadata_history(get_all_related_metadata_history)

        returned = await MetadataHistoryDatabaseFactory(
            engine=test_engine
        ).get_metadata_history_by_metadata_id(metadata_id=38742)
        assert len(returned) == 4

        # assert the first object is the metadata object
        assert returned[0].id == 38742
        assert returned[0].created_by == "test_metadata.py"
        assert returned[0].created_time == datetime(
            2025, 1, 22, 13, 45, 27, tzinfo=timezone.utc
        )
        assert returned[0].updated_by == "another_tester"
        assert returned[0].updated_time == datetime(
            2023, 6, 1, 13, 45, 27, tzinfo=timezone.utc
        )
        assert returned[0].starttime == datetime(
            2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert returned[0].endtime == datetime(
            2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert returned[0].category == MetadataCategory.OBSERVATORY
        assert returned[0].network == "NT"
        assert returned[0].station == "BRW"
        assert returned[0].metadata == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 87.32,
            "elevation": 10,
            "longitude": 202.234,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }

        # assert the metadata_history objects are in order of most recent updated_time
        assert returned[1].id == 3
        assert returned[1].metadata_id == 38742
        assert returned[1].created_by == "test_metadata.py"
        assert returned[1].updated_by == "another_tester"
        assert returned[1].updated_time == datetime(
            2023, 1, 22, 13, 30, 43, tzinfo=timezone.utc
        )
        assert returned[1].starttime == datetime(
            2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert returned[1].endtime == datetime(
            2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert returned[1].category == MetadataCategory.OBSERVATORY
        assert returned[1].network == "NT"
        assert returned[1].station == "BRW"
        assert returned[1].metadata == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 87.32,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }

        assert returned[2].id == 2
        assert returned[2].metadata_id == 38742
        assert returned[2].created_by == "test_metadata.py"
        assert returned[2].updated_by == "tester"
        assert returned[2].updated_time == datetime(
            2023, 1, 21, 3, 30, 43, tzinfo=timezone.utc
        )
        assert returned[2].starttime == datetime(
            2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert returned[2].endtime == datetime(
            2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert returned[2].category == MetadataCategory.OBSERVATORY
        assert returned[2].network == "NT"
        assert returned[2].station == "BRW"
        assert returned[2].metadata == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }

        assert returned[3].id == 1
        assert returned[3].metadata_id == 38742
        assert returned[3].created_by == "test_metadata.py"
        assert returned[3].updated_by == None
        assert returned[3].updated_time == None
        assert returned[3].starttime == datetime(
            2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert returned[3].endtime == datetime(
            2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert returned[3].category == MetadataCategory.OBSERVATORY
        assert returned[3].network == "NT"
        assert returned[3].station == "BRW"
        assert returned[3].metadata == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "12"}],
            "agency": "USGS",
            "latitude": 71.322,
            "elevation": 10,
            "longitude": 203.378,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }


# @pytest.mark.asyncio
async def test_get_metadata_history_by_metadata_id_no_history(
    test_engine, current_metadata_for_related_metadata_history
):
    # clear metadata and MetadataHistoryTable tables
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        history_statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.exec(history_statement)
        test_db_session.commit()

        # save metadata directly to db to have control over the id
        test_db_session.add(current_metadata_for_related_metadata_history)
        test_db_session.commit()

        returned = await MetadataHistoryDatabaseFactory(
            engine=test_engine
        ).get_metadata_history_by_metadata_id(metadata_id=38742)
        assert len(returned) == 1

        # assert the metadata object was returned
        assert returned[0].id == 38742
        assert returned[0].created_by == "test_metadata.py"
        assert returned[0].created_time == datetime(
            2025, 1, 22, 13, 45, 27, tzinfo=timezone.utc
        )
        assert returned[0].updated_by == "another_tester"
        assert returned[0].updated_time == datetime(
            2023, 6, 1, 13, 45, 27, tzinfo=timezone.utc
        )
        assert returned[0].starttime == datetime(
            2022, 2, 3, 22, 43, 27, tzinfo=timezone.utc
        )
        assert returned[0].endtime == datetime(
            2022, 2, 3, 23, 43, 27, tzinfo=timezone.utc
        )
        assert returned[0].category == MetadataCategory.OBSERVATORY
        assert returned[0].network == "NT"
        assert returned[0].station == "BRW"
        assert returned[0].metadata == {
            "id": "BRW",
            "name": "Barrow",
            "marks": [{"name": "m", "azimuth": "1"}],
            "piers": [{"name": "p", "correction": "22"}],
            "agency": "USGS",
            "latitude": 87.32,
            "elevation": 10,
            "longitude": 202.234,
            "agency_name": "United States Geological Survey (USGS)",
            "electronics": ["1"],
            "theodolites": ["12"],
            "declination_base": 10589,
            "sensor_orientation": "HDZF",
        }


@pytest.mark.asyncio
async def test_get_metadata_history_by_metadata_id_no_metadata(test_engine):
    # clear metadata and metadata_history tables
    with Session(test_engine) as test_db_session:
        statement = delete(MetadataTable)
        history_statement = delete(MetadataHistoryTable)
        test_db_session.exec(statement)
        test_db_session.exec(history_statement)
        test_db_session.commit()

        try:
            await MetadataHistoryDatabaseFactory(
                engine=test_engine
            ).get_metadata_history_by_metadata_id(metadata_id=38742)
        except Exception as e:
            assert str(e) == "metadata not found"
