import os
import uuid
import json
import pytest

from sqlmodel import Session, delete, select

from geomagio.api.db.models.session import session
from geomagio.api.db.factories.SessionDatabaseFactory import SessionDatabaseFactory
from geomagio.api.secure.encryption import get_fernet


@pytest.mark.asyncio
async def test_encryption(test_engine, valid_session):
    with Session(test_engine) as test_db_session:
        # clear test session table
        statement = delete(session)
        test_db_session.exec(statement)
        test_db_session.commit()

        session_id = "randomstring"
        await SessionDatabaseFactory(engine=test_engine).save_session(
            session_id=session_id, data=valid_session
        )

        # get session directly from database to make sure its encrypted
        statement = select(session).where(session.session_id == session_id)
        results = test_db_session.exec(statement)
        user_session = results.one()

        # decrypt the data
        encryption = get_fernet()
        decrypted_data = encryption.decrypt(user_session.data.encode("utf8"))

        # assert encryption is working as expected
        returned_session = json.loads(decrypted_data)
        assert returned_session == valid_session


@pytest.mark.asyncio
async def test_get_session_no_session(test_engine):
    try:
        await SessionDatabaseFactory(engine=test_engine).get_session(
            session_id="idonotexist"
        )
    except Exception as e:
        assert e is not None


@pytest.mark.asyncio
async def test_save_new_session(
    test_engine, valid_session, valid_token, valid_userinfo
):
    with Session(test_engine) as test_db_session:
        # clear test session table
        statement = delete(session)
        test_db_session.exec(statement)
        test_db_session.commit()

    session_id = "randomstring"
    await SessionDatabaseFactory(engine=test_engine).save_session(
        session_id=session_id, data=valid_session
    )

    # get the newly created session to make sure it's saved correctly and encryption is working as expected
    user_session = await SessionDatabaseFactory(engine=test_engine).get_session(
        session_id=session_id
    )
    assert user_session["token"] == valid_token
    assert user_session["user"] == valid_userinfo


@pytest.mark.asyncio
async def test_update_current_session(
    test_engine, valid_session, valid_token, valid_userinfo
):
    with Session(test_engine) as test_db_session:
        # clear test session table
        statement = delete(session)
        test_db_session.exec(statement)
        test_db_session.commit()

    # create an initial session to later update
    session_id = "sessiontoupdate"
    await SessionDatabaseFactory(engine=test_engine).save_session(
        session_id=session_id, data=valid_session
    )

    # get the newly created session to make sure it's saved correctly and encryption is working as expected
    user_session = await SessionDatabaseFactory(engine=test_engine).get_session(
        session_id=session_id
    )
    assert user_session["token"] == valid_token
    assert user_session["user"] == valid_userinfo
    assert user_session["user"]["groups"] == ["group1", "group2"]

    # change the userinfo and save it
    updated_session_data = valid_session
    updated_session_data["user"]["groups"] = []
    await SessionDatabaseFactory(engine=test_engine).save_session(
        session_id=session_id, data=updated_session_data
    )

    # get the updated session and assert the group info was changed as expected
    updated_session = await SessionDatabaseFactory(engine=test_engine).get_session(
        session_id=session_id
    )
    assert updated_session["token"] == valid_token
    assert updated_session["user"]["email"] == "test_email"
    assert updated_session["user"]["groups"] == []


@pytest.mark.asyncio
async def test_delete_session(test_engine, valid_session):
    with Session(test_engine) as test_db_session:
        # clear test session table
        statement = delete(session)
        test_db_session.exec(statement)
        test_db_session.commit()

        # create an initial session to later delete
        session_id = "sessiontodelete"
        await SessionDatabaseFactory(engine=test_engine).save_session(
            session_id=session_id, data=valid_session
        )

        # delete session
        await SessionDatabaseFactory(engine=test_engine).delete_session(
            session_id=session_id
        )

        # try to get session to verify
        second_statement = select(session).where(session.session_id == session_id)
        results = test_db_session.exec(second_statement)
        user_session = results.all()

        assert len(user_session) == 0


@pytest.mark.asyncio
async def test_delete_non_existing(test_engine):
    session_id = "idonotexist"
    # this should not error
    await SessionDatabaseFactory(engine=test_engine).delete_session(
        session_id=session_id
    )
