import pytest
import httpx

from unittest.mock import ANY
from fastapi.testclient import TestClient
from authlib.integrations.base_client.errors import MismatchingStateError

from geomagio.api.secure.secure_app import app
from geomagio.api.secure.login_routes import User
from geomagio.api.db.factories.SessionDatabaseFactory import SessionDatabaseFactory

client = TestClient(app)


@pytest.mark.asyncio
async def test_authorization_valid(valid_session, valid_token, valid_userinfo, mocker):
    mock_authorize = mocker.AsyncMock(return_value=valid_token)
    mocker.patch(
        "geomagio.api.secure.login_routes.oauth.openid.authorize_access_token",
        mock_authorize,
    )

    mock_userinfo = mocker.AsyncMock(return_value=valid_userinfo)
    mocker.patch(
        "geomagio.api.secure.login_routes.oauth.openid.userinfo", mock_userinfo
    )

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    response = client.get("/authorize")
    assert response.status_code == 200

    mock_save_session.assert_called_once_with(session_id=ANY, data=valid_session)


@pytest.mark.asyncio
async def test_authorization_invalid_authorize_access_token():
    # assert that requests directly to GET /authorize cause an exception
    with pytest.raises(MismatchingStateError) as err:
        client.get("/authorize")


@pytest.mark.asyncio
async def test_authorization_invalid_authorize_access_token():
    # assert that requests directly to GET /authorize cause an exception
    with pytest.raises(MismatchingStateError) as err:
        client.get("/authorize")

    assert (
        str(err.value)
        == "mismatching_state: CSRF Warning! State not equal in request and response."
    )


@pytest.mark.asyncio
async def test_login_redirects(mocker):
    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    mock_redirect = mocker.AsyncMock(return_value={"status_code": 302})
    mocker.patch(
        "geomagio.api.secure.login_routes.oauth.openid.authorize_redirect",
        mock_redirect,
    )

    header = "https://testserver/ws/secure/metadata/1"
    client.get("/login", headers={"Referer": header})

    # assert oauth is called with the correct redirect url
    mock_redirect.assert_called_once_with(
        ANY, "https://testserver/ws/secure/authorize", prompt="login"
    )

    # assert session is saved with correct referrer header
    mock_save_session.assert_called_once_with(
        session_id=ANY, data={"after_authorize_redirect": header}
    )


@pytest.mark.asyncio
async def test_logout(valid_session, mocker):
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_delete_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "delete_session", mock_delete_session)

    client.get(url="/logout", headers={"Cookie": "PHPSESSID=valid_session_id"})

    mock_get_session.assert_called_once_with(session_id="valid_session_id")
    mock_delete_session.assert_called_once_with(session_id="valid_session_id")


@pytest.mark.asyncio
async def test_user_with_valid_session(valid_token, valid_userinfo, mocker):
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    response = client.get(url="/user", headers={"Cookie": "PHPSESSID=valid_session_id"})

    assert response.status_code == 200
    mock_get_session.assert_called_once_with(session_id="valid_session_id")

    user = User(**response.json())
    assert user.email == "test_email"
    assert user.sub == 1234
    assert user.groups == ["group1", "group2"]


@pytest.mark.asyncio
async def test_user_with_no_session_and_valid_token(
    valid_token, valid_userinfo, valid_gitlab_user, mocker
):
    valid_user_response = httpx.Response(status_code=200, json=valid_gitlab_user)

    valid_groups = [
        {
            "id": 1,
            "full_path": "group1",
        },
        {"id": 2, "full_path": "group2"},
    ]

    valid_groups_response = httpx.Response(status_code=200, json=valid_groups)

    mock_gitlab_request = mocker.AsyncMock(
        side_effect=[
            valid_user_response,
            valid_groups_response,
            httpx.Response(status_code=200, json={}),
        ]
    )
    mocker.patch("httpx.AsyncClient.get", mock_gitlab_request)

    valid_session = {"token": valid_token, "user": valid_userinfo}

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    response = client.get(url="/user", headers={"Authorization": "valid_gitlab_token"})

    session_user = {
        "email": "test_email",
        "sub": 1234,
        "groups": ["group1", "group2"],
        "name": "test_user",
        "nickname": "test_user_nickname",
        "picture": "http://picture_url",
    }

    assert response.status_code == 200
    mock_save_session.assert_called_once_with(
        session_id=ANY, data={"user": session_user}
    )

    user = User(**response.json())
    assert user.email == "test_email"
    assert user.sub == 1234
    assert user.groups == ["group1", "group2"]


@pytest.mark.asyncio
async def test_user_with_no_session(mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    # do not pass in cookie
    response = client.get(url="/user")
    assert response.status_code == 401

    mock_get_session.assert_not_called()


@pytest.mark.asyncio
async def test_user_with_no_session_and_invalid_token(mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_gitlab_request = mocker.AsyncMock(return_value=None)
    mocker.patch(
        "geomagio.api.secure.login_routes.get_gitlab_user", mock_gitlab_request
    )

    # do not pass in cookie
    response = client.get(
        url="/user", headers={"Authorization": "invalid_gitlab_token"}
    )
    assert response.status_code == 401

    mock_get_session.assert_not_called()
    mock_gitlab_request.assert_called_once_with(token="invalid_gitlab_token")


@pytest.mark.asyncio
async def test_user_invalid_session(mocker):
    # mock invalid session. this is created when users GET /metadata without logging in
    invalid_session = {
        "redirect_uri": "redirect_uri",
        "nonce": "nonce_str",
        "url": "test_url",
    }
    mock_get_session = mocker.AsyncMock(return_value=invalid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    response = client.get(
        url="/user", headers={"Cookie": "PHPSESSID=invalid_session_id"}
    )

    assert response.status_code == 401
    mock_get_session.assert_called_once_with(session_id="invalid_session_id")
