import pytest
import json
import httpx

from unittest.mock import ANY
from fastapi.testclient import TestClient
from obspy import UTCDateTime

from geomagio.api.secure.secure_app import app
from geomagio.api.db.models.metadata import Metadata
from geomagio.api.db.factories.MetadataDatabaseFactory import MetadataDatabaseFactory
from geomagio.api.db.factories.MetadataHistoryDatabaseFactory import (
    MetadataHistoryDatabaseFactory,
)
from geomagio.api.db.factories.SessionDatabaseFactory import SessionDatabaseFactory

client = TestClient(app)


@pytest.mark.asyncio
async def test_create_metadata_valid_session(
    valid_session, valid_reading_metadata, mocker
):
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    # create mock response for MetadataDatabaseFactory().create_metadata
    valid_reading_metadata["id"] = 7204
    created_metadata = Metadata(**valid_reading_metadata)
    created_metadata = Metadata.model_validate(created_metadata)

    mock_save_metadata = mocker.AsyncMock(return_value=created_metadata)
    mocker.patch.object(MetadataDatabaseFactory, "create_metadata", mock_save_metadata)

    body = json.dumps(valid_reading_metadata)

    response = client.post(
        url="/metadata", headers={"Cookie": "PHPSESSID=valid_session_id"}, content=body
    )

    assert response.status_code == 201
    mock_get_session.assert_called_once_with(session_id="valid_session_id")

    # assert the response matches the object returned by MetadataDatabaseFactory().create_metadata
    response_body = response.json()
    assert response_body["id"] == 7204
    assert response_body["category"] == "reading"
    assert response_body["starttime"] == UTCDateTime("2024-11-07T12:12:12Z")
    assert response_body["endtime"] == UTCDateTime("2024-11-07T20:20:20Z")
    assert response_body["network"] == "NT"
    assert response_body["station"] == "BOU"
    assert response_body["status"] == "new"
    assert response_body["priority"] == 1
    assert response_body["metadata"] == created_metadata.metadata


@pytest.mark.asyncio
async def test_create_metadata_valid_session_invalid_user_input(valid_session, mocker):
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    invalid_user_input = {
        "created_by": "test_user",
        "starttime": "January 28th, 2025 08:00:00",
        "endtime": "2024-11-07T20:20:20Z",
        "network": "NT",
        "station": "BOU",
        "category": "reading",
        "data_valid": True,
    }

    body = json.dumps(invalid_user_input)

    response = client.post(
        url="/metadata", headers={"Cookie": "PHPSESSID=valid_session_id"}, content=body
    )

    response_body = response.json()

    assert response.status_code == 422
    mock_get_session.assert_called_once_with(session_id="valid_session_id")

    assert (
        response_body["detail"][0]["msg"]
        == "Input should be an instance of UTCDateTime"
    )
    assert (
        response_body["detail"][1]["msg"]
        == "Value error, Invalid time type. See obspy UTCDateTime for more information."
    )


@pytest.mark.asyncio
async def test_create_metadata_valid_auth_token(
    valid_token, valid_userinfo, valid_gitlab_user, valid_reading_metadata, mocker
):
    valid_user_response = httpx.Response(status_code=200, json=valid_gitlab_user)

    valid_groups = [
        {
            "id": 1,
            "full_path": "group1",
        },
        {"id": 2, "full_path": "group2"},
    ]

    valid_groups_response = httpx.Response(status_code=200, json=valid_groups)

    mock_gitlab_request = mocker.AsyncMock(
        side_effect=[
            valid_user_response,
            valid_groups_response,
            httpx.Response(status_code=200, json={}),
        ]
    )
    mocker.patch("httpx.AsyncClient.get", mock_gitlab_request)

    valid_session = {"token": valid_token, "user": valid_userinfo}

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    # create mock response for MetadataDatabaseFactory().create_metadata
    valid_reading_metadata["id"] = 7204
    created_metadata = Metadata(**valid_reading_metadata)
    created_metadata = Metadata.model_validate(created_metadata)
    mock_save_metadata = mocker.AsyncMock(return_value=created_metadata)
    mocker.patch.object(MetadataDatabaseFactory, "create_metadata", mock_save_metadata)

    body = json.dumps(valid_reading_metadata)

    response = client.post(
        url="/metadata", headers={"Authorization": "auth_token"}, content=body
    )

    session_user = {
        "email": "test_email",
        "sub": 1234,
        "groups": ["group1", "group2"],
        "name": "test_user",
        "nickname": "test_user_nickname",
        "picture": "http://picture_url",
    }

    assert response.status_code == 201
    mock_save_session.assert_called_once_with(
        session_id=ANY, data={"user": session_user}
    )

    returned_metadata = Metadata(**response.json())
    assert returned_metadata.id == 7204
    assert returned_metadata.category == "reading"
    assert returned_metadata.starttime == UTCDateTime("2024-11-07T12:12:12Z")
    assert returned_metadata.endtime == UTCDateTime("2024-11-07T20:20:20Z")
    assert returned_metadata.network == "NT"
    assert returned_metadata.station == "BOU"
    assert returned_metadata.status == "new"
    assert returned_metadata.priority == 1


@pytest.mark.asyncio
async def test_create_metadata_no_session_or_token(valid_reading_metadata, mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    body = json.dumps(valid_reading_metadata)

    # do not pass in cookie or auth
    response = client.post(url="/metadata", content=body)
    assert response.status_code == 401

    mock_get_session.assert_not_called()


@pytest.mark.asyncio
async def test_create_metadata_with_no_session_and_invalid_token(mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_gitlab_request = mocker.AsyncMock(return_value=None)
    mocker.patch(
        "geomagio.api.secure.login_routes.get_gitlab_user", mock_gitlab_request
    )

    response = client.post(
        url="/metadata", headers={"Authorization": "invalid_gitlab_token"}
    )
    assert response.status_code == 401

    mock_get_session.assert_not_called()
    mock_gitlab_request.assert_called_once_with(token="invalid_gitlab_token")


@pytest.mark.asyncio
async def test_create_metadata_invalid_session(mocker):
    # mock invalid session. this is created when users GET /metadata without logging in
    invalid_session = {
        "redirect_uri": "redirect_uri",
        "nonce": "nonce_str",
        "url": "test_url",
    }
    mock_get_session = mocker.AsyncMock(return_value=invalid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    response = client.post(
        url="/metadata", headers={"Cookie": "PHPSESSID=invalid_session_id"}
    )

    assert response.status_code == 401
    mock_get_session.assert_called_once_with(session_id="invalid_session_id")


@pytest.mark.asyncio
async def test_update_metadata(
    valid_token, valid_userinfo, instrument_metadata, mocker
):
    valid_userinfo["groups"] = ["ghsc/geomag/operations/roles/reviewer"]
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    # mock response for MetadataDatabaseFactory().update_metadata
    instrument_metadata.id = 7204
    instrument_metadata.updated_by = "test_user"
    instrument_metadata.updated_time = UTCDateTime.now()
    instrument_metadata.metadata = {
        "type": "FGE",
        "channels": {
            "U": [{"channel": "U_Volt", "offset": 0, "scale": 313.2}],
            "V": [{"channel": "V_Volt", "offset": 0, "scale": 312.3}],
            "W": [{"channel": "W_Volt", "offset": 0, "scale": 312.0}],
        },
        "electronics": {
            "serial": "E0542",
            "x-scale": 313.2,
            "y-scale": 312.3,
            "z-scale": 312.0,
            "temperature-scale": 0.01,
        },
        "sensor": {
            "serial": "S0419",
            "x-constant": 36958,
            "y-constant": 36849,
            "z-constant": 36810,
        },
    }
    instrument_metadata = Metadata.model_validate(instrument_metadata)
    mock_update = mocker.AsyncMock(return_value=instrument_metadata)
    mocker.patch.object(MetadataDatabaseFactory, "update_metadata", mock_update)

    valid_user_input = {
        "id": 7204,
        "metadata": {
            "type": "FGE",
            "channels": {
                "U": [{"channel": "U_Volt", "offset": 0, "scale": 313.2}],
                "V": [{"channel": "V_Volt", "offset": 0, "scale": 312.3}],
                "W": [{"channel": "W_Volt", "offset": 0, "scale": 312.0}],
            },
            "electronics": {
                "serial": "E0542",
                "x-scale": 313.2,
                "y-scale": 312.3,
                "z-scale": 312.0,
                "temperature-scale": 0.01,
            },
            "sensor": {
                "serial": "S0419",
                "x-constant": 36958,
                "y-constant": 36849,
                "z-constant": 36810,
            },
        },
    }

    body = json.dumps(valid_user_input)

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=valid_session_id"},
        content=body,
    )
    assert response.status_code == 200
    mock_get_session.assert_called_once_with(session_id="valid_session_id")

    # assert the response matches the object returned by MetadataDatabaseFactory().update_metadata
    response_body = response.json()
    assert response_body["id"] == 7204
    assert response_body["updated_by"] == "test_user"
    assert response_body["category"] == "instrument"
    assert response_body["network"] == "NT"
    assert response_body["station"] == "BDT"
    assert response_body["status"] == "new"
    assert response_body["priority"] == 1
    assert response_body["metadata"] == instrument_metadata.metadata


@pytest.mark.asyncio
async def test_update_metadata_invalid_input(valid_token, valid_userinfo, mocker):
    valid_userinfo["groups"] = ["ghsc/geomag/operations/roles/reviewer"]
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    invalid_user_input = {
        "id": 7204,
        "starttime": "12:20:32",
    }

    body = json.dumps(invalid_user_input)

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=valid_session_id"},
        content=body,
    )
    assert response.status_code == 422
    mock_get_session.assert_called_once_with(session_id="valid_session_id")

    response_body = response.json()

    assert (
        response_body["detail"][0]["msg"]
        == "Input should be an instance of UTCDateTime"
    )
    assert (
        response_body["detail"][1]["msg"]
        == "Value error, Invalid time type. See obspy UTCDateTime for more information."
    )


@pytest.mark.asyncio
async def test_update_metadata_valid_auth_token(
    valid_token,
    valid_userinfo,
    valid_gitlab_user,
    instrument_metadata,
    mocker,
):
    valid_user_response = httpx.Response(status_code=200, json=valid_gitlab_user)

    valid_groups = [
        {
            "id": 1,
            "full_path": "group1",
        },
        {"id": 2, "full_path": "ghsc/geomag/operations/roles/reviewer"},
    ]

    valid_groups_response = httpx.Response(status_code=200, json=valid_groups)

    mock_gitlab_request = mocker.AsyncMock(
        side_effect=[
            valid_user_response,
            valid_groups_response,
            httpx.Response(status_code=200, json={}),
        ]
    )
    mocker.patch("httpx.AsyncClient.get", mock_gitlab_request)

    valid_session = {"token": valid_token, "user": valid_userinfo}

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    # mock response for MetadataDatabaseFactory().update_metadata
    instrument_metadata.id = 7204
    instrument_metadata.updated_by = "test_user"
    instrument_metadata.updated_time = UTCDateTime.now()
    instrument_metadata.metadata = {
        "type": "FGE",
        "channels": {
            "U": [{"channel": "U_Volt", "offset": 0, "scale": 313.2}],
            "V": [{"channel": "V_Volt", "offset": 0, "scale": 312.3}],
            "W": [{"channel": "W_Volt", "offset": 0, "scale": 312.0}],
        },
        "electronics": {
            "serial": "E0542",
            "x-scale": 313.2,
            "y-scale": 312.3,
            "z-scale": 312.0,
            "temperature-scale": 0.01,
        },
        "sensor": {
            "serial": "S0419",
            "x-constant": 36958,
            "y-constant": 36849,
            "z-constant": 36810,
        },
    }
    instrument_metadata = Metadata.model_validate(instrument_metadata)
    mock_update = mocker.AsyncMock(return_value=instrument_metadata)
    mocker.patch.object(MetadataDatabaseFactory, "update_metadata", mock_update)

    valid_user_input = {
        "id": 7204,
        "metadata": {
            "type": "FGE",
            "channels": {
                "U": [{"channel": "U_Volt", "offset": 0, "scale": 313.2}],
                "V": [{"channel": "V_Volt", "offset": 0, "scale": 312.3}],
                "W": [{"channel": "W_Volt", "offset": 0, "scale": 312.0}],
            },
            "electronics": {
                "serial": "E0542",
                "x-scale": 313.2,
                "y-scale": 312.3,
                "z-scale": 312.0,
                "temperature-scale": 0.01,
            },
            "sensor": {
                "serial": "S0419",
                "x-constant": 36958,
                "y-constant": 36849,
                "z-constant": 36810,
            },
        },
    }

    body = json.dumps(valid_user_input)

    response = client.put(
        url="/metadata/1234", headers={"Authorization": "auth_token"}, content=body
    )

    session_user = {
        "email": "test_email",
        "sub": 1234,
        "groups": ["group1", "ghsc/geomag/operations/roles/reviewer"],
        "name": "test_user",
        "nickname": "test_user_nickname",
        "picture": "http://picture_url",
    }

    assert response.status_code == 200
    mock_save_session.assert_called_once_with(
        session_id=ANY, data={"user": session_user}
    )

    # assert the response matches the object returned by MetadataDatabaseFactory().update_metadata
    response_body = response.json()
    assert response_body["id"] == 7204
    assert response_body["updated_by"] == "test_user"
    assert response_body["category"] == "instrument"
    assert response_body["network"] == "NT"
    assert response_body["station"] == "BDT"
    assert response_body["status"] == "new"
    assert response_body["priority"] == 1
    assert response_body["metadata"] == instrument_metadata.metadata


@pytest.mark.asyncio
async def test_update_metadata_no_valid_group(
    valid_session, valid_reading_metadata, mocker
):
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=valid_session_id"},
        content=body,
    )
    assert response.status_code == 403
    mock_get_session.assert_called_once_with(session_id="valid_session_id")


@pytest.mark.asyncio
async def test_update_metadata_no_session(
    valid_session, valid_reading_metadata, mocker
):
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.put(url="/metadata/1234", content=body)
    assert response.status_code == 401

    mock_get_session.assert_not_called()


@pytest.mark.asyncio
async def test_update_metadata_invalid_session(mocker, valid_reading_metadata):
    # mock invalid session. this is created when users GET /metadata without logging in
    invalid_session = {
        "redirect_uri": "redirect_uri",
        "nonce": "nonce_str",
        "url": "test_url",
    }
    mock_get_session = mocker.AsyncMock(return_value=invalid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=invalid_session_id"},
        content=body,
    )

    assert response.status_code == 401
    mock_get_session.assert_called_once_with(session_id="invalid_session_id")


@pytest.mark.asyncio
async def test_update_metadata_with_no_session_and_invalid_token(
    mocker, valid_reading_metadata
):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_gitlab_request = mocker.AsyncMock(return_value=None)
    mocker.patch(
        "geomagio.api.secure.login_routes.get_gitlab_user", mock_gitlab_request
    )

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234",
        headers={"Authorization": "invalid_gitlab_token"},
        content=body,
    )
    assert response.status_code == 401

    mock_get_session.assert_not_called()
    mock_gitlab_request.assert_called_once_with(token="invalid_gitlab_token")


@pytest.mark.asyncio
async def test_update_metadata_no_existing_metadata(
    valid_userinfo, valid_token, mocker
):
    valid_userinfo["groups"] = ["ghsc/geomag/operations/roles/reviewer"]
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionDatabaseFactory, "save_session", mock_save_session)

    mock_update = mocker.AsyncMock(side_effect=ValueError("metadata not found"))
    mocker.patch.object(MetadataDatabaseFactory, "update_metadata", mock_update)

    body = json.dumps({"id": "1234", "category": "instrument"})

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=valid_session_id"},
        content=body,
    )
    assert response.status_code == 404
    mock_get_session.assert_called_once_with(session_id="valid_session_id")


@pytest.mark.asyncio
async def test_get_metadata(instrument_metadata, metadata_with_datetimes, mocker):
    # mock return multiple metadata
    instrument_metadata.id = 1242
    instrument_metadata.station = "BOU"
    metadata_with_datetimes.id = 821
    all_instrument_metadata = [instrument_metadata, metadata_with_datetimes]
    mock_get = mocker.AsyncMock(return_value=all_instrument_metadata)
    mocker.patch.object(MetadataDatabaseFactory, "get_metadata", mock_get)

    response = client.get(url="/metadata?station=BOU&category=instrument&network=NT")
    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 2

    for metadata in response_body:
        assert metadata["station"] == "BOU"
        assert metadata["category"] == "instrument"
        assert metadata["network"] == "NT"

        # assert nested metadata is returned as "metadata" and not "metadata_" to the user
        assert metadata["metadata"] != None
        assert metadata.get("metadata_") == None


@pytest.mark.asyncio
async def test_get_metadata_return_single(instrument_metadata, mocker):
    # mock return single metadata
    instrument_metadata.id = 1242
    all_instrument_metadata = [instrument_metadata]
    mock_get = mocker.AsyncMock(return_value=all_instrument_metadata)
    mocker.patch.object(MetadataDatabaseFactory, "get_metadata", mock_get)

    response = client.get(url="/metadata?station=BDT&category=instrument&network=NT")
    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 1
    assert response_body[0]["station"] == "BDT"
    assert response_body[0]["category"] == "instrument"
    assert response_body[0]["network"] == "NT"

    # assert nested metadata is returned as "metadata" and not "metadata_" to the user
    assert response_body[0]["metadata"] != None
    assert response_body[0].get("metadata_") == None


@pytest.mark.asyncio
async def test_get_metadata_return_empty(mocker):
    # mock return empty list

    mock_get = mocker.AsyncMock(return_value=[])
    mocker.patch.object(MetadataDatabaseFactory, "get_metadata", mock_get)

    response = client.get(url="/metadata?station=BDT&category=instrument&network=NT")
    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 0


@pytest.mark.asyncio
async def test_get_metadata_invalid_input():
    response = client.get(
        url="/metadata?station=BDT&category=instrument&starttime=1:12:24T8:8:43"
    )
    assert response.status_code == 422

    response_body = response.json()
    assert (
        response_body["detail"][0]["msg"]
        == "Input should be an instance of UTCDateTime"
    )
    assert (
        response_body["detail"][1]["msg"]
        == "Value error, Invalid time type. See obspy UTCDateTime for more information."
    )


@pytest.mark.asyncio
async def test_get_metadata_by_id(instrument_metadata, mocker):
    instrument_metadata.id = 1242
    mock_get = mocker.AsyncMock(return_value=instrument_metadata)
    mocker.patch.object(MetadataDatabaseFactory, "get_metadata_by_id", mock_get)

    response = client.get(url="/metadata/1242")
    assert response.status_code == 200

    response_body = response.json()
    assert response_body["station"] == "BDT"
    assert response_body["category"] == "instrument"
    assert response_body["network"] == "NT"

    # assert nested metadata is returned as "metadata" and not "metadata_" to the user
    assert response_body["metadata"] == {
        "type": "FGE",
        "channels": {
            "U": [{"channel": "U_Volt", "offset": 0, "scale": 313.2}],
            "V": [{"channel": "V_Volt", "offset": 0, "scale": 312.3}],
            "W": [{"channel": "W_Volt", "offset": 0, "scale": 312.0}],
        },
        "electronics": {
            "serial": "E0542",
            "x-scale": 313.2,
            "y-scale": 312.3,
            "z-scale": 312.0,
            "temperature-scale": 0.01,
        },
        "sensor": {
            "serial": "S0419",
            "x-constant": 36958,
            "y-constant": 36849,
            "z-constant": 36811,
        },
    }
    assert response_body.get("metadata_") == None


@pytest.mark.asyncio
async def test_get_metadata_by_id_no_metadata_found(mocker):
    mock_get = mocker.AsyncMock(return_value=None)
    mocker.patch.object(MetadataDatabaseFactory, "get_metadata_by_id", mock_get)

    response = client.get(url="/metadata/1242")
    assert response.status_code == 404
    response_body = response.json()
    assert response_body["detail"] == "metadata not found"


@pytest.mark.asyncio
async def test_get_metadata_by_id_invalid_id():
    response = client.get(url="/metadata/828cskje9w8e")
    assert response.status_code == 422

    response_body = response.json()
    assert (
        response_body["detail"][0]["msg"]
        == "Input should be a valid integer, unable to parse string as an integer"
    )


@pytest.mark.asyncio
async def test_get_metadata_history(
    formatted_metadata_history_1, formatted_metadata_history_2, mocker
):
    # mock return multiple metadata_history
    mock_get = mocker.AsyncMock(
        return_value=[formatted_metadata_history_1, formatted_metadata_history_2]
    )
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history", mock_get
    )

    response = client.get(
        url="/metadata/history?station=BRW&category=observatory&network=NT"
    )
    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 2

    for metadata_history in response_body:
        assert metadata_history["metadata_id"] == 38746
        assert metadata_history["station"] == "BRW"
        assert metadata_history["category"] == "observatory"
        assert metadata_history["network"] == "NT"

        # assert nested metadata is returned as "metadata" and not "metadata_" to the user
        assert metadata_history["metadata"] != None
        assert metadata_history.get("metadata_") == None


@pytest.mark.asyncio
async def test_get_metadata_history_return_single(formatted_metadata_history_1, mocker):
    # mock return single MetadataHistoryTable
    mock_get = mocker.AsyncMock(return_value=[formatted_metadata_history_1])
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history", mock_get
    )

    response = client.get(
        url="/metadata/history?station=BRW&category=observatory&network=NT"
    )
    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 1

    assert response_body[0]["metadata_id"] == 38746
    assert response_body[0]["station"] == "BRW"
    assert response_body[0]["category"] == "observatory"
    assert response_body[0]["network"] == "NT"

    # assert nested metadata is returned as "metadata" and not "metadata_" to the user
    assert response_body[0]["metadata"] != None
    assert response_body[0].get("metadata_") == None


@pytest.mark.asyncio
async def test_get_metadata_history_return_empty(mocker):
    # mock return empty

    mock_get = mocker.AsyncMock(return_value=[])
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history", mock_get
    )

    response = client.get(
        url="/metadata/history?station=BRW&category=observatory&network=NT"
    )
    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 0


@pytest.mark.asyncio
async def test_get_metadata_history_invalid_input():
    response = client.get(
        url="/metadata/history?station=BRW&category=observatory&network=NT&starttime=today"
    )
    assert response.status_code == 422

    response_body = response.json()
    assert (
        response_body["detail"][0]["msg"]
        == "Input should be an instance of UTCDateTime"
    )
    assert (
        response_body["detail"][1]["msg"]
        == "Value error, Invalid time type. See obspy UTCDateTime for more information."
    )


@pytest.mark.asyncio
async def test_get_metadata_history_by_id(formatted_metadata_history_1, mocker):
    mock_get = mocker.AsyncMock(return_value=formatted_metadata_history_1)
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history_by_id", mock_get
    )

    response = client.get(url="/metadata/history/124221")
    assert response.status_code == 200

    response_body = response.json()
    assert response_body["metadata_id"] == 38746
    assert response_body["station"] == "BRW"
    assert response_body["category"] == "observatory"
    assert response_body["network"] == "NT"

    # assert nested metadata is returned as "metadata" and not "metadata_" to the user
    assert response_body["metadata"] == {
        "id": "BRW",
        "name": "Barrow",
        "marks": [{"name": "m", "azimuth": "1"}],
        "piers": [{"name": "p", "correction": "12"}],
        "agency": "USGS",
        "latitude": 71.322,
        "elevation": 10,
        "longitude": 203.378,
        "agency_name": "United States Geological Survey (USGS)",
        "electronics": ["1"],
        "theodolites": ["12"],
        "declination_base": 10589,
        "sensor_orientation": "HDZF",
    }
    assert response_body.get("metadata_") == None


@pytest.mark.asyncio
async def test_get_metadata_history_by_id_no_metadata_history_found(mocker):
    mock_get = mocker.AsyncMock(return_value=None)
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history_by_id", mock_get
    )

    response = client.get(url="/metadata/history/34253")
    assert response.status_code == 404
    response_body = response.json()
    assert response_body["detail"] == "metadata_history not found"


@pytest.mark.asyncio
async def test_get_metadata_history_by_id_invalid_id():
    response = client.get(url="/metadata/history/8239453knsoi23yr")
    assert response.status_code == 422

    response_body = response.json()
    assert (
        response_body["detail"][0]["msg"]
        == "Input should be a valid integer, unable to parse string as an integer"
    )


@pytest.mark.asyncio
async def test_get_metadata_history_by_metadata_id(
    get_all_formatted_metadata_history, mocker
):
    mock_get = mocker.AsyncMock(return_value=get_all_formatted_metadata_history)
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history_by_metadata_id", mock_get
    )

    response = client.get(url="/metadata/12532/history")

    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 4
    assert response_body[0]["id"] == 38746
    assert response_body[1]["metadata_id"] == 38746
    assert response_body[2]["metadata_id"] == 38746
    assert response_body[3]["metadata_id"] == 38746


@pytest.mark.asyncio
async def test_get_metadata_history_by_metadata_id_only_metadata(
    formatted_metadata_for_formatted_metadata_history, mocker
):
    mock_get = mocker.AsyncMock(
        return_value=[formatted_metadata_for_formatted_metadata_history]
    )
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history_by_metadata_id", mock_get
    )

    response = client.get(url="/metadata/12532/history")

    assert response.status_code == 200

    response_body = response.json()
    assert len(response_body) == 1
    assert response_body[0]["id"] == 38746


@pytest.mark.asyncio
async def test_get_metadata_history_by_metadata_id_not_found(mocker):
    mock_get = mocker.AsyncMock(return_value=None)
    mocker.patch.object(
        MetadataHistoryDatabaseFactory, "get_metadata_history_by_metadata_id", mock_get
    )

    response = client.get(url="/metadata/3824598235973/history")

    assert response.status_code == 404

    response_body = response.json()
    assert response_body["detail"] == "metadata not found"
