import datetime
from fastapi import Depends
from fastapi.testclient import TestClient
from numpy.testing import assert_equal
from obspy import UTCDateTime
import pytest

from geomagio.api.ws.ws_app import app
from geomagio.api.ws.data import get_data_query
from geomagio.api.ws.DataApiQuery import (
    DataApiQuery,
    OutputFormat,
    SamplingPeriod,
    DataType,
)


@pytest.fixture(scope="module")
def test_client():
    @app.get("/query/", response_model=DataApiQuery)
    def get_query(query: DataApiQuery = Depends(get_data_query)):
        return query

    client = TestClient(app)
    yield client


def test_get_data_query(test_client):
    """test.api_test.ws_test.data_test.test_get_data_query()"""
    response = test_client.get(
        "/query/?id=BOU&starttime=2020-09-01T00:00:01&elements=X,Y,Z,F&type=variation&sampling_period=60&format=iaga2002"
    )
    query = DataApiQuery(**response.json())
    assert_equal(query.id, "BOU")
    assert_equal(query.starttime, UTCDateTime("2020-09-01T00:00:01"))
    assert_equal(query.endtime, UTCDateTime("2020-09-02T00:00:00.999"))
    assert_equal(query.elements, ["X", "Y", "Z", "F"])
    assert_equal(query.sampling_period, SamplingPeriod.MINUTE)
    assert_equal(query.format, "iaga2002")
    assert_equal(query.data_type, "variation")


def test_get_data_query_no_starttime(test_client):
    """test.api_test.ws_test.data_test.test_get_data_query_no_starttime()"""
    response = test_client.get("/query/?id=BOU&elements=X,Y,Z,F")
    query = DataApiQuery(**response.json())
    assert_equal(query.id, "BOU")

    # expect starttime to default to start of current day
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    expected = UTCDateTime(year=now.year, month=now.month, day=now.day)

    assert_equal(query.starttime, expected)
    # expect endtime to default to 1 day after starttiime
    assert_equal(query.endtime, expected + (86400 - 0.001))


async def test_get_data_query_extra_params(test_client):
    with pytest.raises(ValueError) as error:
        response = test_client.get(
            "/query/?id=BOU&starttime=2020-09-01T00:00:01&elements=X,Y,Z,F&type=variation&sampling_period=60&format=iaga2002&location=R1&network=NT"
        )
        DataApiQuery(**response.json())
        assert error.match("Invalid query parameter(s): location, network")


def test_get_data_query_bad_params(test_client):
    """test.api_test.ws_test.data_test.test_get_data_query_bad_params()"""
    with pytest.raises(ValueError) as error:
        response = test_client.get(
            "/query/?id=BOU&startime=2020-09-01T00:00:01&elements=X,Y,Z,F&data_type=variation&sampling_period=60&format=iaga2002"
        )
        DataApiQuery(**response.json())
        assert error.match == "Invalid query parameter(s): startime, data_type"
